# -*- coding: utf-8 -*-
"""
Created on 2020/9/14
@project: SPAIC
@filename: Torch_Backend
@author: Hong Chaofei
@contact: hongchf@gmail.com
@description:
"""
from .Backend import Backend, backends
import torch
import numpy as np
# from torch import fx
#from torch.nn import Module, Parameter
import torch.nn.functional as fn
import torch.nn as nn
from typing import Tuple, Dict, Callable
from collections import defaultdict
from torch.cuda.amp import autocast
import torch.distributed
import threading

class Torch_Engine(torch.nn.Module):
    def __init__(self, graph_operations):
        super(Torch_Engine, self).__init__()
        self._graph_operations = graph_operations

    def forward(self, variables: Dict[str,torch.Tensor]):
        temp_dict = dict()
        update_dict = dict()
        reduce_dict = dict()

        for op in self._graph_operations:
            # for inputs
            inputs = []
            for var in op.input:
                if var[0] == 'variables_dict':
                    inputs.append(variables[var[1]])
                elif var[0] == 'temp_dict':
                    inputs.append(temp_dict[var[1]])
                elif var[0] == 'update_dict':
                    inputs.append(update_dict[var[1]])
                elif var[0] == 'reduce_dict':
                    inputs.append(reduce_dict[var[1]])
            # compute the operation
            result = op.func(*inputs)
            if len(op.output) == 1: result = [result]
            # assign the result variables
            for ind, var in enumerate(op.output):
                if var[0] == 'temp_dict':
                    temp_dict[var[1]] = result[ind]
                elif var[0] == 'update_dict':
                    update_dict[var[1]] = result[ind]
                elif var[0] == 'reduce_dict':
                    if var[1] in reduce_dict:
                        reduce_dict[var[1]].append(result[ind])
                    else:
                        reduce_dict[var[1]] = [result[ind]]

        return update_dict



class Torch_Backend(Backend):
    backend_name = 'pytorch'

    def __init__(self, device='cpu'):
        super(Torch_Backend, self).__init__()
        self.device = device if isinstance(device,list) else [device]
        self.device0 = self.device[0]
        self.device_count = len(self.device)
        self.data_type = torch.float32
        self.debug_data = []
        self.nograd_decorator = torch.no_grad()
        self.enablegrad_decorator = torch.enable_grad()
        pass

    def build(self):
        from torch import fx
        # self._graph_var_dicts = {'variables_dict': self._variables, 'temp_dict': dict(), 'update_dict': dict(),
        #                          'reduce_dict': dict()}
        # self._graph_var_dicts['temp_dict']['example_temp_dict_pytorch_datatype'] = torch.empty(1)
        # self._graph_var_dicts['update_dict']['example_temp_dict_pytorch_datatype'] = torch.empty(1)
        # self._graph_var_dicts['reduce_dict']['example_temp_dict_pytorch_datatype'] = torch.empty(1)
        #
        #self.update_step = jit.trace(self.update_step)
        self.engine = Torch_Engine(self._graph_operations)
        self.engine = fx.symbolic_trace(self.engine)

        # self.graph_update_step = torch.jit.script(self.engine)
        self.graph_update_step = self.engine
        # print(self.engine.code)

    def build_graph(self):

        for key, value in self._InitVariables_dict.items():
            if isinstance(value, torch.Tensor):
                value = value.to(self.device0)
                self._InitVariables_dict[key] = value

        super(Torch_Backend, self).build_graph()

    def remove_tensor(self, inputs):
        if not torch.is_tensor(inputs[0]):
            device = inputs[1].device
        else:
            device = inputs[0].device
        for ind,t in enumerate(inputs):
            if torch.is_tensor(t):
                inputs[ind] = t.to(device)

    def is_insert(self,inputs):
        if len(inputs) == 0:
            return False
        elif len(set(inputs)) < len(inputs):
            return True

    def move_compute_and_assign_tensors(self,op):
        inputs = []
        for var in op.input:
            if torch.is_tensor(var.value):
                var.value = var.value.to(op.place)
                inputs.append(var.value)
            elif isinstance(var.value, list):
                for ind,x in enumerate(var.value):
                    if torch.is_tensor(x):
                        var.value[ind] = x.to(op.place)
                inputs.append(var.value)
            else:
                inputs.append(var.value)
        result = op.func(*inputs)
        if len(op.output) == 1: result = [result]
        for ind, var in enumerate(op.output):
            var.value = result[ind]

    def graph_update_step_multigpu(self):
        def _worker(*tuple):
            grad_enabled, autocast_enabled = torch.is_grad_enabled(), torch.is_autocast_enabled()
            for group in sorted(tuple):
                op = self._graph_operations[group]
                # with torch.cuda.device(self.device[index]), autocast(enabled=autocast_enabled):
                self.move_compute_and_assign_tensors(op)

        if self.partition:
            if self.partition == 'multithread':
                threads = [threading.Thread(target=_worker,args=group) for group in self.groups]
                for thread in threads:
                    thread.start()
                for thread in threads:
                    thread.join()
            if self.partition == 'thread1':
                for group in self.groups:
                    for i in sorted(group):
                        op = self._graph_operations[i]
                        self.move_compute_and_assign_tensors(op)
            for op in self.isolate:
                op = self._graph_operations[op]
                self.move_compute_and_assign_tensors(op)
        else:
            for group in self.groups:
                for op in group:
                    self.move_compute_and_assign_tensors(op)

    def to_nograd_func(self, func):
        return self.nograd_decorator(func)

    def to_grad_func(self, func):
        return self.enablegrad_decorator(func)

    #  As of now, autograd support floating point Tensor types ( half, float, double and bfloat16) and complex Tensor types (cfloat, cdouble).
    def add_backend_variable(self, module, name, shape, value=None, grad=False, is_sparse=False, init=None, init_param=None, prefer_device=None):
        '''
        Parameters
        ----------
        name
        shape
        value
        init
        Returns
        -------
        '''
        l = len(self._parameters_dict)
        if prefer_device != None:
            device0 = self.device[prefer_device]
        else:
            device0 = self.device[l % self.device_count]

        if init_param is None:
            init_param = dict()
        if init is not None:
            # self._variables[name] = self.init_param(grad, init)
            data = torch.empty(shape, dtype=self.data_type, device=device0, requires_grad=grad)
            init = init.lower()
            if init in self.param_init_operate.keys():
                self._variables[name] = self.param_init_operate[init](data, **init_param)
            else:
                raise ValueError("No initialize method: %s in param_init_operate" % init)
        elif value is not None:
        # if value is not None:
            if hasattr(value, "__len__"):
                if (value.shape == torch.Size([1,]) or value.shape == torch.Size([])) and isinstance(value, torch.Tensor):
                    self._variables[name] = (value.to(self.device0) * torch.ones(shape, dtype=self.data_type, device=self.device0)).clone()
                    self._variables[name].requires_grad = grad
                elif tuple(value.shape) != tuple(shape):
                    raise ValueError("Value is not scalar and the shape of Value is not equal to shape")
                # add a sparse matrices with all dimensions greater than 2
                elif is_sparse:
                    i = np.nonzero(value)
                    v = value[i]

                    # Index for sparse matrix
                    sparse_index = name + '_sparse_index'
                    self._variables[sparse_index] = torch.LongTensor(i).to(device=self.device0)
                    self._InitVariables_dict[sparse_index] = self._variables[sparse_index]

                    # Value for sparse matrix
                    sparse_value = name + '_sparse_value'
                    if init is not None:
                        # self._variables[sparse_value] = self.init_param(True, init)
                        data = torch.empty(shape, dtype=self.data_type, device=self.device0, requires_grad=True)
                        self._variables[sparse_value] = self.param_init_operate[init](data, **init_param)
                    else:
                        self._variables[sparse_value] = torch.tensor(v, dtype=self.data_type, requires_grad=True, device=self.device0)
                    self._parameters_dict[sparse_value] = self._variables[sparse_value]

                    # The shape of sparse matrix
                    sparse_shape = name + '_sparse_shape'
                    self._variables[sparse_shape] = torch.Size(shape)
                    self._InitVariables_dict[sparse_shape] = self._variables[sparse_shape]

                    # Sparse matrix
                    self._variables[name] = torch.sparse.FloatTensor(self._variables[sparse_index], self._variables[sparse_value], self._variables[sparse_shape])
                else:
                    # add a non sparse matrices with all dimensions greater than 2
                    if init is not None:
                        data = torch.empty(shape, dtype=self.data_type, device=self.device0, requires_grad=grad)
                        init = init.lower()
                        if init in self.param_init_operate.keys():
                            self._variables[name] = self.param_init_operate[init](data, **init_param)
                        else:
                            raise ValueError("No initialize method: %s in param_init_operate" % init)
                    else:
                        if isinstance(value, torch.Tensor):
                            # device0 = random.choice(self.device)
                            self._variables[name] = value.clone().detach().to(device0)
                        else:
                            self._variables[name] = torch.tensor(value, dtype=self.data_type, device=device0,
                                                                   requires_grad=grad)

            elif len(shape) == 0:
                # add constant
                self._variables[name] = torch.tensor(value, dtype=self.data_type, device=self.device0, requires_grad=grad)

            else:
                # add a matrix through constant
                if init is not None:
                    # self._variables[name] = self.init_param(grad, init)
                    data = value*torch.ones(shape, dtype=self.data_type, device=device0, requires_grad=grad)
                    init = init.lower()
                    if init in self.param_init_operate.keys():
                        self._variables[name] = self.param_init_operate[init](data, **init_param)
                    else:
                        raise ValueError("No initialize method: %s in param_init_operate" % init)
                else:
                    # add a matrix through constant
                    self._variables[name] = (
                                value * torch.ones(shape, dtype=self.data_type, device=self.device0)).clone()
                    self._variables[name].requires_grad = grad
        return self._variables[name]

    def set_variable_value(self, name, value, is_parameter):
        if is_parameter:
            assert name in self._parameters_dict
            assert self._parameters_dict[name].shape == value.shape
            if not isinstance(value, torch.Tensor):
                value = torch.tensor(value, dtype=self._parameters_dict[name].dtype, device=self._parameters_dict[name].device)
            with torch.no_grad():
                self._parameters_dict[name].data = value
        else:
            assert name in self._InitVariables_dict
            assert self._backend._InitVariables_dict[name].shape == value.shape
            if not isinstance(value, torch.Tensor):
                value = torch.tensor(value, dtype=self._InitVariables_dict[name].dtype, device=self._InitVariables_dict[name].device)
            with torch.no_grad():
                self._InitVariables_dict[name].data = value


    # def init_param(self, grad, *init):
    #     if init[0] in self.param_init_operate:
    #         init_op = self.param_init_operate[init[0]]
    #     else:
    #         raise ValueError("No init operate %s in param_init_operate" % init[0])
    #     inputs = []
    #     shape = init[1]
    #     data = torch.empty(shape, dtype=self.data_type, device=self.device, requires_grad=grad)
    #     inputs.append(data)
    #
    #     for var in init[2:]:
    #         inputs.append(var)
    #     return init_op(*inputs)

    def sparse_to_dense(self, index_name, value_name, shape_name):
        return torch.sparse.FloatTensor(self._variables[index_name], self._variables[value_name], self._variables[shape_name])

    def get_str(self, level):
        return level*' ' + 'torch_backend'

    def threshold(self, x, v_th):
        return torch.gt(x, v_th).type(self.data_type)

    def reset(self, v, o):
        return o.eq(0)*v

    def cat(self, x, dim=1):
        return torch.cat(x, dim)

    def stack(self, x, dim=1):       # 在指定维度dim上连接（concatenate）若干个张量。
        try:
            return torch.stack(x, dim)
        except:
            # patch for SLIF 2[O]
            for ii in range(len(x)):
                if x[ii].dim() ==2:
                    tmp = torch.zeros_like(x[ii])
                    tmp = torch.stack([x[ii], tmp], dim=1)
                    x[ii] = tmp
            return torch.stack(x, dim)

    def reduce_sum(self, x, *dim):
        if len(dim)==0:
            dim = 1
        return torch.sum(x, dim=dim)

    def index_select(self, x,  indices, dim=1):
        return torch.index_select(x, dim=dim, index=indices)

    def permute(self, x, permute_dim):
        return x.permute(permute_dim)

    def view(self, x, view_dim):

        x = x.contiguous().view(view_dim)
        return x


    def scatter(self, x, indices):
        return torch.scatter(x, dim=0, index=indices)

    def conv1d(self, x, kernel):
        return torch.conv1d(x, kernel)

    def conv_trans1d(self, x, kernel, bias=None):
        return torch.conv_transpose1d(x, kernel, bias)


    def conv_2d(self, x, kernel, stride, padding, dilation, groups, padding_mode='constant'):
        if x.dim() == kernel.dim() + 1:
            xshape = list(x.shape)
            xshape[0] = xshape[0]*xshape[1]
            extend_size = xshape[1]
            xshape.pop(1)
            out = fn.conv2d(x.reshape(xshape), kernel, stride=stride, padding=padding, dilation=dilation, groups=groups, padding_mode=padding_mode)
            outshape = list(out.shape)
            outshape[0] = outshape[0]//extend_size
            outshape.insert(1, extend_size)
            return out.view(outshape)
        else:
            return fn.conv2d(x, kernel, stride=stride, padding=padding, dilation=dilation, groups=groups)

    def conv_2d_complex(self, x, kernel, stride, padding, dilation, groups, beta, delay=None):
        if x.dtype.is_complex:
            if delay is not None:
                d_delay = delay/self.dt
                d_delay = torch.ceil(d_delay) - d_delay
                x = beta ** (x.imag+d_delay) * (x.real * (0 + 1.0j))
            else:
                x = beta ** x.imag * (x.real * (0 + 1.0j))
        else:
            x = x*(0+1.0j)
        real = fn.conv2d(x.real, kernel, stride=stride, padding=padding, dilation=dilation, groups=groups)
        imag = fn.conv2d(x.imag, kernel, stride=stride, padding=padding, dilation=dilation, groups=groups)
        return torch.complex(real, imag)

    def conv_trans2d(self, x, kernel, stride=1, padding=0, dilation=0, groups=1):
        return torch.conv_transpose2d(x, kernel,stride=stride, padding=padding, dilation=dilation, groups=groups)



    def conv_max_pool2d(self, x, kernel, pool_kernel, stride, pool_stride, padding, pool_padding, dilation, groups):
        return fn.max_pool2d(fn.conv2d(x, kernel, stride=stride, padding=padding,
                             dilation=dilation, groups=groups), kernel_size=pool_kernel,
                             stride=pool_stride, padding=pool_padding)

        # return fn.conv2d(fn.max_pool2d(x, int(max_kernel[0])), kernel, stride=int(stride), padding=int(padding), dilation=int(dilation), groups=int(groups))

    def conv_avg_pool2d(self, x, kernel, pool_kernel, stride, pool_stride, padding, pool_padding, dilation, groups):
        return fn.avg_pool2d(fn.conv2d(x, kernel, stride=stride, padding=padding,
                             dilation=dilation, groups=groups), kernel_size=pool_kernel,
                             stride=pool_stride, padding=pool_padding)
    def conv_add_bias(self, x, bias):
        bias_t = bias.repeat(x.shape[-2], x.shape[-1], 1).permute(2, 1, 0)
        return x+bias_t
    def max_pool2d(self, x, pool_kernel, pool_stride, pool_padding):
        return fn.max_pool2d(x, kernel_size=pool_kernel, stride=pool_stride, padding=pool_padding)
    def post_max_pool2d_complex(self, x, pool_kernel, pool_stride, pool_padding):
        pool_imag, pool_index = fn.max_pool2d(x.imag, kernel_size=pool_kernel, return_indices=True,
                                              stride=pool_stride, padding=pool_padding)
        x_shape = x.shape
        pool_shape = pool_index.shape
        pool_real = torch.gather(x.real.view(x_shape[0], x_shape[1], -1), dim=-1,
                                 index=pool_index.view(x_shape[0], x_shape[1],-1)).view(pool_shape[0], pool_shape[1],
                                                                                        pool_shape[2], pool_shape[3])


        return torch.complex(real=pool_real, imag=pool_imag)

    def avg_pool2d(self, x, pool_kernel, pool_stride, pool_padding):
        return fn.avg_pool2d(x, kernel_size=pool_kernel, stride=pool_stride, padding=pool_padding)

    def batchnorm2d(self, x, num_features):
        # 该实现方式忽略了running_mean 和 running_var
        # 当 batch_size 较小时，在推理阶段的统计特性就会和全局统计特性有着较大偏差，从而导致糟糕的效果，这种情况下推荐使用
        # SPAIC 的 Module 模块实现 batchnorm。
        device = x.device
        bn_2d = torch.nn.BatchNorm2d(num_features).to(device=device)
        return bn_2d(x)

    def dropout(self, x, p, inplace=False):
        return fn.dropout(x, p=p, inplace=inplace)

    def reshape_mat_mult(self, A, X):

        if A.dim() == 4:
            (batchsize, outchannels, H, W) = A.shape
            A = A.view(batchsize, -1)
        elif A.dim() == 5:
            (batchsize, extend, outchannels, H, W) = A.shape
            A = A.view(batchsize, extend, -1)

        return torch.matmul(A, X.permute(1, 0))

    def im2col_indices(self, x, kh, kw, padding, stride):
        return fn.unfold(x, (kh, kw), padding=padding, stride=stride)

    def conv2d_flatten(self, x):
        return x.view(x.shape[0], x.shape[1], -1)

    def feature_map_flatten(self, x):
        return x.view(x.shape[0], -1)

    def add(self, x, y):

        return x + y

    def minus(self, x, y):
        return x - y

    def div(self, x, y):
        return torch.div(x, y)

    def relu(self, x):
        return torch.relu(x)

    def sigmoid(self, x):
        return torch.sigmoid(x)

    def mat_mult_weight(self, A, X):
        '''
        Parameters
        ----------
        A--->preGroup:input
        X--->postGroup:weight
        Returns
        -------
        '''

        X = X.permute(1, 0)
        return torch.matmul(A, X)


    def mat_mult_weight_complex(self, A, X, beta, delay=None):
        '''
        Parameters
        ----------
        A--->preGroup:input
        X--->postGroup:weight
        beta---> postGroup:beta_complex
        Returns
        -------
        '''
        if A.dtype.is_complex:
            beta = beta.unsqueeze(-1)
            if delay is not None:
                A = A.permute(0,2,1)
                real = A.real
                imag = A.imag
                d_delay = delay.unsqueeze(0)/self.dt
                d_delay = torch.ceil(d_delay) - d_delay
                O = beta ** (imag+d_delay) * (real * (0 + 1.0j))
            else:
                A = A.unsqueeze(-2)
                real = A.real
                imag = A.imag
                O = beta ** imag * (real * (0 + 1.0j))
            # if torch.any(torch.isnan(O)):
            #     print("real:", real)
            #     print("real:", real)
            #     print("imag:", imag)
            #     print("O:", O)
            #     raise ValueError(" nan mat_mult_complex error")
            return torch.sum(O * X, dim=-1)
        elif delay is not None:
            A = A.permute(0, 2, 1)
            return torch.sum(A * X, dim=-1)*(0.0+1.0j)
        else:
            X = X.permute(1, 0)
            Out = torch.matmul(A.to(X.dtype), X)
            Out = Out*(0.0+1.0j)
            # if torch.any(torch.isnan(Out)):
            #     print("input:", A)
            #     print("weight:", X)
            #     print("Out:", Out)
            #     raise ValueError(" nan mat_mult_complex error")
            return Out

    def mat_mult_weight_2complex(self, A, X, beta, delay=None):
        if A.dtype.is_complex:
            A = A.unsqueeze(-2).unsqueeze(-1)
            beta = beta.unsqueeze(-1)
            real = A.real
            imag = A.imag
            O = torch.sum((beta ** imag * (real * (0 + 1.0j))) * X, dim=-2)
            return O
        else:
            A = A.unsqueeze(-2).unsqueeze(-1)
            O = torch.sum(A*X, dim=-2)
            return O

    def mat_mult_pre(self, A, X):
        '''
        Parameters
        ----------
        A--->preGroup:input
        X--->postGroup:weight
        Returns
        -------
        '''
        A = A.permute(1, 0)
        return torch.matmul(A, X)

    def mat_mult(self, A, X):
        '''
        Parameters
        ----------
        A--->preGroup:input
        X--->postGroup:weight
        Returns
        -------
        '''
        return torch.matmul(A, X)

    def bmm(self, A, X):
        '''
        Parameters
        ----------
        A---> postGroup
        X---> preGroup
        Returns
        -------
        '''
        return torch.bmm(A, X)

    def ger(self, A, X):
        '''
        Parameters
        ----------
        A---> postGroup
        X---> preGroup
        Returns
        -------
        '''
        return torch.ger(A, X)

    def sparse_mat_mult_weight(self, A, X):
        '''
       Parameters
       ----------
       A--->preGroup:sparseWeight(post, pre)
       X--->postGroup:input(batch, pre)
       Returns
       -------
       '''
        X = X.permute(1, 0)
        result = torch.sparse.mm(A, X)
        result = result.permute(1, 0)
        return result

    def var_mult(self, A, X):
        return A * X

    def mult_sum_weight(self, A, X):
        # X = X.permute(1, 0)
        # A = A.permute(0, 2, 1)
        return torch.sum(A * X, dim=-1)

    def mat_linear(self, A, X, b):
        return torch.matmul(A, X) + b

    def var_linear(self, A, X, b):

        return A*X+b

    def unsqueeze(self, X, dim):
        return torch.unsqueeze(X, dim)

    def to_numpy(self, data: torch.Tensor):
        return data.detach().cpu().numpy()

    def to_tensor(self, data):
        if isinstance(data, torch.Tensor):
            return data.to(torch.float).to(self.device)
        else:
            return torch.tensor(data, dtype=torch.float, device=self.device)
    def upsample(self,x, scale):
        return torch.nn.functional.interpolate(x, scale_factor=scale, mode='nearest')

    def exp(self, x):
        return torch.exp(x)


    def clamp_(self, data, min, max):
        with torch.no_grad():
            data.clamp_(min, max)


    def clamp_max_(self, data, max):
        with torch.no_grad():
            data.clamp_max_(max)



    def clamp_min_(self, data, min):
        with torch.no_grad():
            data.clamp_min_(min)


    def uniform(self, data, a=-0.0, b=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            a(float): the lower bound of the uniform distribution
            b(float): the upper bound of the uniform distribution
        Returns:
            torch.nn.init.uniform_(data, a=0.0, b=1.0)
        '''
        return torch.nn.init.uniform_(data, a, b)


    def normal(self, data, mean=0.0, std=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            mean(float): the mean of the normal distribution
            std(float): the standard deviation of the normal distribution
        Returns:
            torch.nn.init.normal_(data, mean=0.0, std=1.0)
        '''
        return torch.nn.init.normal_(data, mean, std)



    def xavier_normal(self, data, gain=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            gain: an optional scaling factor
        Returns:
            torch.nn.init.xavier_normal_(data, gain=1.0)
        '''
        return torch.nn.init.xavier_normal_(data, gain)


    def xavier_uniform(self, data, gain=1.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            gain: an optional scaling factor
        Returns:
            torch.nn.init.xavier_uniform_(data, gain=1.0)
        '''
        return torch.nn.init.xavier_uniform_(data, gain)


    def kaiming_normal(self, data, a=0, mode='fan_in', nonlinearity='leaky_relu'):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            a: the negative slope of the rectifier used after this layer (only used with 'leaky_relu')
            mode: either 'fan_in' (default) or 'fan_out'. Choosing 'fan_in' preserves the magnitude of the variance of the weights in the forward pass. Choosing 'fan_out' preserves the magnitudes in the backwards pass.
            nonlinearity: the non-linear function (nn.functional name), recommended to use only with 'relu' or 'leaky_relu' (default).
        Returns:
            torch.nn.init.kaiming_normal_(data, a=0, mode='fan_in', nonlinearity='leaky_relu')
        '''
        return torch.nn.init.kaiming_normal_(data, a, mode, nonlinearity)


    def kaiming_uniform(self, data, a=0, mode='fan_in', nonlinearity='leaky_relu'):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            a: the negative slope of the rectifier used after this layer (only used with 'leaky_relu')
            mode: either 'fan_in' (default) or 'fan_out'. Choosing 'fan_in' preserves the magnitude of the variance of the weights in the forward pass. Choosing 'fan_out' preserves the magnitudes in the backwards pass.
            nonlinearity: the non-linear function (nn.functional name), recommended to use only with 'relu' or 'leaky_relu' (default).
        Returns:
            torch.nn.init.kaiming_uniform_(data, a=0, mode='fan_in', nonlinearity='leaky_relu')
        '''
        return torch.nn.init.kaiming_uniform_(data, a, mode, nonlinearity)

    def constant(self, data, constant_value=0.0):
        '''
        Args:
            data(tensor): an n-dimensional torch.Tensor
            constant_value(float): the value to fill the tensor with
        Returns:
            torch.nn.init.constant_(data, constant_value)
        '''
        return torch.nn.init.constant_(data, constant_value)

    def sparse(self, data, sparsity=0.1, std=0.01):
        '''
        Args:
            data(tensor): an n-dimensional `torch.Tensor`
            sparsity(float): The fraction of elements in each column to be set to zero
            std(float): the standard deviation of the normal distribution used to generate
            the non-zero values
        Returns:
            torch.nn.init.sparse_(data, sparsity, std)
        '''
        return torch.nn.init.sparse_(data, sparsity, std)

    def weight_norm(self, weight, amp):
        w_norm = torch.norm(weight, p=2, dim=1, keepdim=True)
        # print(amp.item(), w_norm.item())
        return weight*amp/w_norm

    #TODO: THis "TO" should be named to_device
    def to(self, x, device):
        return x.to(device)

    def sin(self, x):
        return torch.sin(x)

    def cos(self, x):
        return torch.cos(x)

    def tan(self, x):
        return torch.tan(x)

    def log(self, x):
        return torch.log(x)

    def log2(self, x):
        return torch.log2(x)

    def log10(self, x):
        return torch.log10(x)

    # def reset(self, x, v_reset, u_reset, spike):
    #
    #     # if hasattr(x, "__len__"):
    #     #     if x.shape != spike.shape:
    #     #         raise ValueError("%s and %s do not match" % (x.shape, spike.shape))
    #     mask = torch.eq(spike, 1)
    #     x[mask] = v_reset
    #     x[mask] += u_reset
    #     return x

    # def izh_v(self, v, u, psp):
    #     v = v+self.dt*(0.04*v*v+5*v+140-u+psp)
    #     return v
    #
    # def izh_u(self, a, b, v, u):
    #     u = u+self.dt*a*(b*v-u)
    #     return u


backends[Torch_Backend.backend_name] = Torch_Backend





# test = Torch_Backend()
# th = test.basic_operate['threshold']
# print(th(-1.0))


# -*- coding: utf-8 -*-
"""
Created on 2021/8/9
@project: main.py
@filename: OnlineSpikePlot
@author: Hong Chaofei
@contact: hongchf@gmail.com

@description:
在训练过程中实现放电脉冲栅图更新
"""
import spaic
import matplotlib.pyplot as plt
from multiprocessing import Queue, Process, Lock
from matplotlib.animation import FuncAnimation


class SpikePlot(spaic.BaseModule):

    def __init__(self, spike_monitors = [], plot_interval=1):
        super(SpikePlot, self).__init__()
        self.spike_monitors = spike_monitors
        self.interval = plot_interval
        self.draw_proc = None
        self.queue = Queue(maxsize=10)
        self.mutex = Lock()





# -*- coding: utf-8 -*-
import os
import matplotlib.pyplot as plt
import numpy as np

import json
import pickle

from spaic.IO.utils import load_kp_data, save_kp_feature, save_mfcc_feature, load_mfcc_data, load_aedat_v3, un_tar, \
    create_same_directory_structure, integrate_events_file_to_frames_file_by_fixed_frames_number, \
    integrate_events_file_to_frames_file_by_fixed_duration
from concurrent.futures import ThreadPoolExecutor
import multiprocessing
import time

class Dataset(object):
    r"""
    All datasets that represent a map from keys to data samples should subclass it.
    All subclasses should overwrite :meth:`__getitem__`, supporting fetching a data
    sample for a given key.
    Subclasses should also overwrite :meth:`__len__`, which is expected to return
    the size of the sample dataset.
    """

    def __init__(self, **kwargs):
        super().__init__()

    # 根据索引返回数据内容和标签
    def __getitem__(self, index):
        raise NotImplementedError

    # 返回数据集大小
    def __len__(self):
        raise NotImplementedError

class CustomDataset(Dataset):
    r"""
    自定义数据集：
    个人采集的实值数据
    """

    def __init__(self, data=None, label=None):
        super().__init__()
        data_type = type(data)
        label_type = type(label)
        assert data_type is list or data_type is np.ndarray, "The type of data should be list or np.ndarray"
        assert label_type is list or label_type is np.ndarray, "The type of label should be list or np.ndarray"

        self.data = data
        self.label = label

    def __getitem__(self, index):
        data = np.float32(self.data[index])
        label = np.int64(self.label[index])
        return data, label

    def __len__(self):
        return len(self.data)

class CustomSpikeDataset(Dataset):
    r"""
    自定义数据集：
    编码后的脉冲数据（仅支持[spike_time, neuron_ids]表示）
    """

    def __init__(self, spike_times=None, neuron_ids=None, label=None):
        super().__init__()
        # The shape of spike_times should be [sample_num, spikes_times_num]
        spike_times_type = type(spike_times)
        neuron_ids_type = type(neuron_ids)
        label_type = type(label)
        assert spike_times_type is list or spike_times_type is np.ndarray, "The type of data should be list or np.ndarray"
        assert neuron_ids_type is list or neuron_ids_type is np.ndarray, "The type of data should be list or np.ndarray"
        assert label_type is list or label_type is np.ndarray, "The type of label should be list or np.ndarray"

        self.spike_times = spike_times
        self.neuron_ids = neuron_ids
        self.label = label

    def __getitem__(self, index):
        spike_time = self.spike_times[index]
        neuron_id = self.neuron_ids[index]
        label = self.label[index]
        spiking = [spike_time, neuron_id]
        return spiking, label

    def __len__(self):
        return len(self.data)


class SpecifiedDataset(Dataset):

    r"""
    labels load from json file
    """
    def __init__(self, image_file, label_file):
        super().__init__()
        # 加载数据集
        label_file = label_file
        img_folder = image_file

        fp = open(label_file, 'r')
        data_dict = json.load(fp)

        # 如果图像数和标签数不匹配说明数据集标注生成有问题，报错提示
        assert len(data_dict['images']) == len(data_dict['labels'])
        num_data = len(data_dict['images'])

        self.filenames = []
        self.labels = []
        self.img_folder = img_folder
        for i in range(num_data):
            self.filenames.append(data_dict['images'][i])
            self.labels.append(data_dict['labels'][i])

    def __getitem__(self, index):
        img_name = np.float32(self.img_folder + self.filenames[index])
        label = np.int64(self.labels[index])
        img = plt.imread(img_name)
        return img, label

    def __len__(self):
        return len(self.filenames)

class cifar10(Dataset):
    files = {
        "train_dataset1": 'data_batch_1',
        "train_dataset2": 'data_batch_2',
        "train_dataset3": 'data_batch_3',
        "train_dataset4": 'data_batch_4',
        "train_dataset5": 'data_batch_5',
        "test_dataset": 'test_batch'
    }

    def __init__(self, root, is_train=True):
        super().__init__()
        self.root = root
        self._is_train = is_train
        self.data = {
            'train_images': [],
            'test_images': [],
            'train_labels': [],
            'test_labels': []
        }
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")
        if self._dataset_exists():
            self._to_numpy_format()
        else:
            raise ValueError(">> Failed to load the set, file not exist. You should download the dataset firstly.")

    def __getitem__(self, index):
        # 数据归一化到[0,1]
        mean = (0.4914, 0.4822, 0.4465)
        std = (0.2023, 0.1994, 0.2010)

        if self._is_train:
            img = (self.data['train_images'][index]/255.0 - mean)/std
            # img = (self.data['train_images'][index])
            img = np.float32(img.transpose(2, 0, 1))
            # img = np.float32(self.data['train_images'][index])
            label = np.int64(self.data['train_labels'][index])
        else:
            img = (self.data['test_images'][index] / 255.0 - mean) / std
            # img = (self.data['test_images'][index])
            img = np.float32(img.transpose(2, 0, 1))
            label = np.int64(self.data['test_labels'][index])

        return img, label

    def __len__(self):
        if self._is_train:
            return len(self.data['train_images'])
        else:
            return len(self.data['test_images'])

    @property
    def dataset_folder(self):
        return os.path.join(self.root, self.__class__.__name__)

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def load_cifar10_batch(self, folder_path, batch_id):

        with open(folder_path + '/data_batch_' + str(batch_id), mode='rb') as file:
            batch = pickle.load(file, encoding='latin1')

        # features and labels
        features = batch['data'].reshape((len(batch['data']), 3, 32, 32)).transpose(0, 2, 3, 1)  # batch['data']
        labels = batch['labels']

        return features, labels

    def _to_numpy_format(self):

        self.data['train_images'], self.data['train_labels'] = self.load_cifar10_batch(self.root, 1)

        for batch_id in range(2, 6):
            features, labels = self.load_cifar10_batch(self.root, batch_id)
            self.data['train_images'] = np.concatenate([self.data['train_images'], features])
            self.data['train_labels'] = np.concatenate([self.data['train_labels'], labels])

        with open(self.root + '/test_batch', mode='rb') as f:
            batch = pickle.load(f, encoding='latin1')
            self.data['test_images'] = batch['data'].reshape((len(batch['data']), 3, 32, 32)).transpose(0, 2, 3, 1)  # batch['data']
            self.data['test_labels'] = batch['labels']

        print(">> Dataset loaded")

    def _dataset_exists(self):
        if os.path.exists(os.path.join(self.root)):
            for file in cifar10.files.values():
                if not os.path.isfile(os.path.join(self.root, file)):
                    return False
            return True
        else:
            return False


class ImageNet(Dataset):
    files = {
        "train_dataset": 'ILSVRC2012_img_train.tar',
        "val_dataset": 'ILSVRC2012_img_val.tar',
        "val_label": 'ILSVRC2012_devkit_t12.tar.gz'
    }

    def __init__(self, root, is_train=True):
        super().__init__()
        self.root = root
        self._is_train = is_train
        self.data = {
            'train_images': [],
            'val_images': [],
            'train_labels': [],
            'val_labels': []
        }
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")
        if self._dataset_exists():
            self.untar_train_tar(ImageNet.files['train_dataset'].split('.')[0])
            val_dir = ImageNet.files['val_dataset'].split('.')[0]
            devkit_dir = ImageNet.files['val_label'].split('.')[0]
            self.move_val_img(val_dir=val_dir, devkit_dir=devkit_dir)

            pass
            # self._to_numpy_format()
        else:
            raise ValueError(">> Failed to load the set, file not exist. You should download the dataset firstly.")

    def __getitem__(self, index):
        # 数据归一化到[0,1]
        if self._is_train:
            img = np.float32(self.data['train_images'][index])
            label = np.int64(self.data['train_labels'][index])
        else:
            img = np.float32(self.data['test_images'][index])
            label = np.int64(self.data['test_labels'][index])

        return img / 255.0, label

    def __len__(self):
        if self._is_train:
            return len(self.data['train_images'])
        else:
            return len(self.data['test_images'])

    @property
    def dataset_folder(self):
        return os.path.join(self.root, self.__class__.__name__)

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def untar_train_tar(self, train_tar):
        """
        untar images from train_tar and save in corresponding folders
        organize like:
        /train
           /n01440764
               images
           /n01443537
               images
            .....
        """
        root, _, files = next(os.walk(os.path.join(self.root, train_tar)))
        for file in files:
            un_tar(os.path.join(root, file), os.path.join(self.root, 'train'))

    def move_val_img(self, val_dir, devkit_dir):
        """
        move val_img to correspongding folders.
        val_id(start from 1) -> ILSVRC_ID(start from 1) -> WIND
        organize like:
        /val
           /n01440764
               images
           /n01443537
               images
            .....
        """
        # load synset, val ground truth and val images list
        from scipy import io
        import shutil
        devkit_dir_name = os.path.join(self.root, devkit_dir)
        synset = io.loadmat(os.path.join(devkit_dir_name, 'ILSVRC2012_devkit_t12', 'data', 'meta.mat'))

        ground_truth = open(os.path.join(devkit_dir_name, 'ILSVRC2012_devkit_t12', 'data', 'ILSVRC2012_validation_ground_truth.txt'))
        lines = ground_truth.readlines()
        labels = [int(line[:-1]) for line in lines]

        val_dir_name = os.path.join(self.root, val_dir)
        root, _, filenames = next(os.walk(val_dir_name))
        for filename in filenames:
            # val image name -> ILSVRC ID -> WIND
            val_id = int(filename.split('.')[0].split('_')[-1])
            ILSVRC_ID = labels[val_id - 1]
            WIND = synset['synsets'][ILSVRC_ID - 1][0][1][0]
            print("val_id:%d, ILSVRC_ID:%d, WIND:%s" % (val_id, ILSVRC_ID, WIND))

            # move val images
            output_dir = os.path.join(self.root, 'val', WIND)
            if os.path.isdir(output_dir):
                pass
            else:
                os.mkdir(output_dir)
            shutil.move(os.path.join(root, filename), os.path.join(output_dir, filename))

    def _dataset_exists(self):
        if os.path.exists(os.path.join(self.root)):
            for file in ImageNet.files.values():
                if os.path.isfile(os.path.join(self.root, file)):
                    # file_name = os.path.join(self.root, file)
                    un_tar(os.path.join(self.root, file), self.root)
                else:
                    return False
            return True
        else:
            return False


class MNIST(Dataset):
    r"""
    A 10-class multi-class classfication
    Args:
    root (string): Root directory of dataset where ``MNIST/processed/training.pt``
            and  ``MNIST/processed/test.pt`` exist.
    is_train (bool, optional): If True, creates dataset from ``training.pt``,
            otherwise from ``test.pt``.
    """
    class_number = 10
    maxNum = 28*28
    resources = [
        "http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz",
        "http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz",
        "http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz",
        "http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz",
    ]

    files = {
        "train_images": 'train-images-idx3-ubyte',
        "test_images": 't10k-images-idx3-ubyte',
        "train_labels": 'train-labels-idx1-ubyte',
        "test_labels": 't10k-labels-idx1-ubyte',
    }

    def __init__(self, root, is_train=True):

        super().__init__()
        self.root = root
        self._is_train = is_train
        self.data = {}
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")
        if self._dataset_exists():
            self._to_numpy_format()
        else:
            self.download()

    def __getitem__(self, index):
        # 数据归一化到[0,1]
        if self._is_train:
            img = np.float32(self.data['train_images'][index])
            label = np.int64(self.data['train_labels'][index])
        else:
            img = np.float32(self.data['test_images'][index])
            label = np.int64(self.data['test_labels'][index])

        return img / 255.0, label

    def __len__(self):
        if self._is_train:
            return len(self.data['train_images'])
        else:
            return len(self.data['test_images'])

    @property
    def dataset_folder(self):
        return os.path.join(self.root, self.__class__.__name__)

    @property
    def class_to_idx(self):
        return {_class: i for i, _class in enumerate(self.classes)}

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def _to_numpy_format(self):
        with open(
                os.path.join(self.root, MNIST.files['train_images']), 'rb'
        ) as f:
            self.data['train_images'] = np.frombuffer(
                f.read(), np.uint8, offset=16
            ).reshape(-1, 28 ** 2)
        with open(
                os.path.join(self.root, MNIST.files['train_labels']), 'rb'
        ) as f:
            self.data['train_labels'] = np.frombuffer(
                f.read(), np.uint8, offset=8
            )
        with open(
                os.path.join(self.root, MNIST.files['test_images']), 'rb'
        ) as f:
            self.data['test_images'] = np.frombuffer(
                f.read(),
                np.uint8,
                offset=16
            ).reshape(-1, 28 ** 2)
        with open(
                os.path.join(self.root, MNIST.files['test_labels']), 'rb'
        ) as f:
            self.data['test_labels'] = np.frombuffer(
                f.read(), np.uint8, offset=8
            )
        print(">> Dataset loaded")

    def _dataset_exists(self):
        if os.path.exists(os.path.join(self.root)):
            for file in MNIST.files.values():
                if not os.path.isfile(os.path.join(self.root, file)):
                    return False
            return True
        else:
            return False

    def download(self):
        if self._dataset_exists():
            print(">> Dataset already exists. ")
            return
        else:
            raise ValueError(">> Failed to load the set, file not exist. You should download the dataset firstly.")
            pass


class FashionMNIST(Dataset):
    r"""
    A 10-class multi-class classfication
    """
    class_number = 10
    maxNum = 28*28
    resources = [
        "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-images-idx3-ubyte.gz",
        "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/train-labels-idx1-ubyte.gz",
        "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-images-idx3-ubyte.gz",
        "http://fashion-mnist.s3-website.eu-central-1.amazonaws.com/t10k-labels-idx1-ubyte.gz",
    ]

    files = {
        "train_images": 'train-images-idx3-ubyte',
        "test_images": 't10k-images-idx3-ubyte',
        "train_labels": 'train-labels-idx1-ubyte',
        "test_labels": 't10k-labels-idx1-ubyte',
    }

    def __init__(self, root, is_train=True):

        super().__init__()
        self.root = root
        self._is_train = is_train
        self.data = {}
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")
        if self._dataset_exists():
            self._to_numpy_format()
        else:
            self.download()

    def __getitem__(self, index):
        # 数据归一化到[0,1]
        if self._is_train:
            img = np.float32(self.data['train_images'][index])
            label = np.int64(self.data['train_labels'][index])
        else:
            img = np.float32(self.data['test_images'][index])
            label = np.int64(self.data['test_labels'][index])

        return img / 255.0, label

    def __len__(self):
        if self._is_train:
            return len(self.data['train_images'])
        else:
            return len(self.data['test_images'])

    @property
    def dataset_folder(self):
        return os.path.join(self.root, self.__class__.__name__)

    @property
    def class_to_idx(self):
        return {_class: i for i, _class in enumerate(self.classes)}

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def _to_numpy_format(self):
        with open(
                os.path.join(self.root, FashionMNIST.files['train_images']), 'rb'
        ) as f:
            self.data['train_images'] = np.frombuffer(
                f.read(), np.uint8, offset=16
            ).reshape(-1, 28 ** 2)
        with open(
                os.path.join(self.root, FashionMNIST.files['train_labels']), 'rb'
        ) as f:
            self.data['train_labels'] = np.frombuffer(
                f.read(), np.uint8, offset=8
            )
        with open(
                os.path.join(self.root, FashionMNIST.files['test_images']), 'rb'
        ) as f:
            self.data['test_images'] = np.frombuffer(
                f.read(),
                np.uint8,
                offset=16
            ).reshape(-1, 28 ** 2)
        with open(
                os.path.join(self.root, FashionMNIST.files['test_labels']), 'rb'
        ) as f:
            self.data['test_labels'] = np.frombuffer(
                f.read(), np.uint8, offset=8
            )
        print(">> Dataset loaded")

    def _dataset_exists(self):
        if os.path.exists(os.path.join(self.root)):
            for file in FashionMNIST.files.values():
                if not os.path.isfile(os.path.join(self.root, file)):
                    return False
            return True
        else:
            return False


class PathMNIST(Dataset):
    r"""
    A 9-class multi-class classfication
    """
    resources = 'https://drive.google.com/drive/folders/1Tl_SP-ffDQg-jDG_EWPlWKgZTmGbvFXU'
    class_number = 9
    maxNum = 28*28*3
    def __init__(self, root, is_train=True):
        super().__init__()
        self.root = root
        self._is_train = is_train
        self.data = {
            'train_images': [],
            'test_images': [],
            'train_labels': [],
            'test_labels': []
        }
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")
        if self._dataset_exists():
            self._to_numpy_format()
        else:
            raise ValueError(">> Failed to load the set, file not exist. You should download the dataset firstly.")

    def __getitem__(self, index):
        if self._is_train:
            img = np.float32(self.data['train_images'][index].flatten() / 255)
            label = np.int64(self.data['train_labels'][index])
            return img, label
        else:
            img = np.float32(self.data['test_images'][index].flatten() / 255)
            label = np.int64(self.data['test_labels'][index])
            return img, label

    def __len__(self, ):
        if self._is_train:
            return len(self.data['train_images'])
        else:
            return len(self.data['test_images'])

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def _to_numpy_format(self):
        unzip_data = np.load(os.path.join(self.root, 'pathmnist.npz'))
        self.data['train_images'] = unzip_data['train_images']
        self.data['test_images'] = unzip_data['test_images']
        self.data['train_labels'] = unzip_data['train_labels'].squeeze()
        self.data['test_labels'] = unzip_data['test_labels'].squeeze()
        return

    def _dataset_exists(self):
        if os.path.isfile(os.path.join(self.root, 'pathmnist.npz')):
            return True
        else:
            return False


class OctMNIST(Dataset):
    r"""
    A 4-class multi-class classfication
    """
    resources = 'https://drive.google.com/drive/folders/1Tl_SP-ffDQg-jDG_EWPlWKgZTmGbvFXU'
    class_number = 4
    maxNum = 28 * 28

    def __init__(self, root, is_train=True):
        super().__init__()
        self.root = root
        self._is_train = is_train
        self.data = {
            'train_images': [],
            'test_images': [],
            'train_labels': [],
            'test_labels': []
        }
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")
        if self._dataset_exists():
            self._to_numpy_format()
        else:
            raise ValueError(">> Failed to load the set, file not exist. You should download the dataset firstly.")

    def __getitem__(self, index):
        if self._is_train:
            img = np.float32(self.data['train_images'][index].flatten() / 255)
            label = np.int64(self.data['train_labels'][index])
            return img, label
        else:
            img = np.float32(self.data['test_images'][index].flatten() / 255)
            label = np.int64(self.data['test_labels'][index])
            return img, label

    def __len__(self, ):
        if self._is_train:
            return len(self.data['train_images'])
        else:
            return len(self.data['test_images'])

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def _to_numpy_format(self):
        unzip_data = np.load(os.path.join(self.root, 'octmnist.npz'))
        self.data['train_images'] = unzip_data['train_images']
        self.data['test_images'] = unzip_data['test_images']
        self.data['train_labels'] = unzip_data['train_labels'].squeeze()
        self.data['test_labels'] = unzip_data['test_labels'].squeeze()
        return

    def _dataset_exists(self):
        if os.path.isfile(os.path.join(self.root, 'octmnist.npz')):
            return True
        else:
            return False


class RWCP10(Dataset):
    r"""
    """
    classes = {
        "ring": 0,
        "whistle1": 1,
        "phone4": 2,
        "cymbals": 3,
        "horn": 4,
        "bells5": 5,
        "buzzer": 6,
        "kara": 7,
        "metal15": 8,
        "bottle1": 9
    }
    class_number = 10

    def __init__(self, root, is_train=True, **kwargs):
        super().__init__()
        self.scale = kwargs.get('scale', 0.1)
        preprocessing = kwargs.get('preprocessing', 'mfcc')
        self.preprocessing = preprocessing.lower()
        self.root = root
        npz_name = kwargs.get('npz_name', ('mfcc_feature', 'kp_feature'))

        self._is_train = is_train
        self.data = {
            'train_audios': [],
            'test_audios': [],
            'train_ids': [],
            'test_ids': [],
            'train_labels': [],
            'test_labels': [],
            'Time': [],
            'neuron_num': []
        }
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")

        if self._dataset_exists():
            self._npz_exists(npz_name)

            if self.npz_name == 'kp_feature.npz':
                self.data = load_kp_data(self.root, self.npz_name)
                self.maxTime = int(np.ceil(self.data['Time'] * self.scale))
                self.maxNum = int(self.data['neuron_num'])

            elif self.npz_name == 'mfcc_feature.npz':
                self.data = load_mfcc_data(self.root, self.npz_name)
                self.maxTime = 50
                self.maxNum = int(self.data['neuron_num'])

            # 如果npz_name不存在
            else:
                if self._classfile_exists():
                    if self.preprocessing == 'kp':
                        self.npz_name = save_kp_feature(root=self.root, npz_name=self.npz_name, sample_rate=16e3,
                                                        class_labels=RWCP10.classes)
                        self.data = load_kp_data(self.root, self.npz_name)
                        self.maxTime = int(np.ceil(self.data['Time'] * self.scale))
                        self.maxNum = int(self.data['neuron_num'])
                    elif self.preprocessing == 'mfcc':
                        self.npz_name = save_mfcc_feature(root=self.root, npz_name=self.npz_name, sample_rate=16e3,
                                                          class_labels=RWCP10.classes)
                        self.data = load_mfcc_data(self.root, self.npz_name)
                        self.maxTime = 50
                        self.maxNum = int(self.data['neuron_num'])
                    else:
                        raise ValueError(">> Wrong preprocessing method. Please select kp or mfcc")

        else:
            raise ValueError(">> Failed to load the set, file not exist. You should download the dataset firstly.")

    def __getitem__(self, index):
        if self._is_train:
            if self.npz_name == 'kp_feature.npz':
                spiking = [self.data['train_audios'][index] * self.scale, self.data['train_ids'][index]]
            else:
                spiking = (self.data['train_audios'][index]).astype(float)
            label = np.int64(self.data['train_labels'][index])
        else:
            if self.npz_name == 'kp_feature.npz':
                spiking = [self.data['test_audios'][index] * self.scale, self.data['test_ids'][index]]
            else:
                spiking = (self.data['test_audios'][index]).astype(float)
            label = np.int64(self.data['test_labels'][index])
        return spiking, label

    def __len__(self):
        if self._is_train:
            return len(self.data['train_audios'])
        else:
            return len(self.data['test_audios'])

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def _dataset_exists(self):
        if os.path.exists(self.root):
            return True
        else:
            return False

    # 判断原始数据是否存在
    def _classfile_exists(self):
        for cls_id in RWCP10.classes.keys():
            if not os.path.isdir(os.path.join(self.root, 'test', cls_id)):
                return False
            if not os.path.isdir(os.path.join(self.root, 'train', cls_id)):
                return False
        return True

    def _npz_exists(self, npz_name):
        file_name = os.listdir(self.root)
        if npz_name == ('mfcc_feature', 'kp_feature'):
            if 'mfcc_feature.npz' in file_name:
                self.npz_name = 'mfcc_feature.npz'
            elif 'kp_feature.npz' in file_name:
                self.npz_name = 'kp_feature.npz'
            else:
                self.npz_name = ''
        else:
            if npz_name in file_name:
                self.npz_name = npz_name
            else:
                self.npz_name = ''


class MNISTVoices(Dataset):
    r"""
    Used to load any type of 0-9 audio dataset
    Class number: 10
    """
    class_number = 10
    classes = {
        "0": 0,
        "1": 1,
        "2": 2,
        "3": 3,
        "4": 4,
        "5": 5,
        "6": 6,
        "7": 7,
        "8": 8,
        "9": 9
    }

    def __init__(self, root, is_train=True, **kwargs):
        super().__init__()
        self.scale = kwargs.get('scale', 0.1)
        preprocessing = kwargs.get('preprocessing', 'mfcc')
        self.preprocessing = preprocessing.lower()
        self.root = root
        npz_name = kwargs.get('npz_name', ('mfcc_feature', 'kp_feature'))

        self._is_train = is_train
        self.data = {
            'train_audios': [],
            'test_audios': [],
            'train_ids': [],
            'test_ids': [],
            'train_labels': [],
            'test_labels': [],
            'Time': [],
            'neuron_num': []
        }
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")

        if self._dataset_exists():
            self._npz_exists(npz_name)

            if self.npz_name == 'kp_feature.npz':
                self.data = load_kp_data(self.root, self.npz_name)
                self.maxTime = int(np.ceil(self.data['Time'] * self.scale))
                self.maxNum = int(self.data['neuron_num'])

            elif self.npz_name == 'mfcc_feature.npz':
                self.data = load_mfcc_data(self.root, self.npz_name)
                self.maxTime = 50
                self.maxNum = int(self.data['neuron_num'])

            # 如果npz_name不存在
            else:
                if self._classfile_exists():
                    if self.preprocessing == 'kp':
                        self.npz_name = save_kp_feature(root=self.root, npz_name=self.npz_name, sample_rate=16e3,
                                                   class_labels=MNISTVoices.classes)
                        self.data = load_kp_data(self.root, self.npz_name)
                        self.maxTime = int(np.ceil(self.data['Time'] * self.scale))
                        self.maxNum = int(self.data['neuron_num'])
                    elif self.preprocessing == 'mfcc':
                        self.npz_name = save_mfcc_feature(root=self.root, npz_name=self.npz_name, sample_rate=16e3,
                                                     class_labels=MNISTVoices.classes)
                        self.data = load_mfcc_data(self.root, self.npz_name)
                        self.maxTime = 50
                        self.maxNum = int(self.data['neuron_num'])
                    else:
                        raise ValueError(">> Wrong preprocessing method. Please select kp or mfcc")

        else:
            raise ValueError(">> Failed to load the set, file not exist. You should download the dataset firstly.")

    def __getitem__(self, index):
        if self._is_train:
            if self.npz_name == 'kp_feature.npz':
                spiking = [self.data['train_audios'][index] * self.scale, self.data['train_ids'][index]]
            else:
                spiking = (self.data['train_audios'][index]).astype(float)
            label = np.int64(self.data['train_labels'][index])
        else:
            if self.npz_name == 'kp_feature.npz':
                spiking = [self.data['test_audios'][index] * self.scale, self.data['test_ids'][index]]
            else:
                spiking = (self.data['test_audios'][index]).astype(float)
            label = np.int64(self.data['test_labels'][index])
        return spiking, label

    def __len__(self):
        if self._is_train:
            return len(self.data['train_audios'])
        else:
            return len(self.data['test_audios'])

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def _dataset_exists(self):
        if os.path.exists(self.root):
            return True
        else:
            return False

    # 判断原始数据是否存在
    def _classfile_exists(self):
        for cls_id in MNISTVoices.classes.keys():
            if not os.path.isdir(os.path.join(self.root, 'test', cls_id)):
                return False
            if not os.path.isdir(os.path.join(self.root, 'train', cls_id)):
                return False
        return True

    def _npz_exists(self, npz_name):
        file_name = os.listdir(self.root)
        if npz_name == ('mfcc_feature', 'kp_feature'):
            if 'mfcc_feature.npz' in file_name:
                self.npz_name = 'mfcc_feature.npz'
            elif 'kp_feature.npz' in file_name:
                self.npz_name = 'kp_feature.npz'
            else:
                self.npz_name = ''
        else:
            if npz_name in file_name:
                self.npz_name = npz_name
            else:
                self.npz_name = ''



class TIDIGITS(Dataset):
    r"""
    Used to load any type of 0-9 and oh audio dataset
    Class number: 11
    """
    class_number = 11
    classes = {
        "zero": 0,
        "one": 1,
        "two": 2,
        "three": 3,
        "four": 4,
        "five": 5,
        "six": 6,
        "seven": 7,
        "eight": 8,
        "nine": 9,
        "oh": 10
    }

    def __init__(self, root, is_train=True, **kwargs):
        super().__init__()
        self.scale = kwargs.get('scale', 0.1)
        preprocessing = kwargs.get('preprocessing', 'mfcc')
        self.preprocessing = preprocessing.lower()
        self.root = root
        npz_name = kwargs.get('npz_name', ('mfcc_feature', 'kp_feature'))

        self._is_train = is_train
        self.data = {
            'train_audios': [],
            'test_audios': [],
            'train_ids': [],
            'test_ids': [],
            'train_labels': [],
            'test_labels': [],
            'Time': [],
            'neuron_num': []
        }
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")

        if self._dataset_exists():
            self._npz_exists(npz_name)

            if self.npz_name == 'kp_feature.npz':
                self.data = load_kp_data(self.root, self.npz_name)
                self.maxTime = int(np.ceil(self.data['Time'] * self.scale))
                self.maxNum = int(self.data['neuron_num'])

            elif self.npz_name == 'mfcc_feature.npz':
                self.data = load_mfcc_data(self.root, self.npz_name)
                self.maxTime = 40
                self.maxNum = int(self.data['neuron_num'])

            # 如果npz_name不存在
            else:
                if self._classfile_exists():
                    if self.preprocessing == 'kp':
                        self.npz_name = save_kp_feature(root=self.root, npz_name=self.npz_name, sample_rate=16e3,
                                                   class_labels=TIDIGITS.classes)
                        self.data = load_kp_data(self.root, self.npz_name)
                        self.maxTime = int(np.ceil(self.data['Time'] * self.scale))
                        self.maxNum = int(self.data['neuron_num'])
                    elif self.preprocessing == 'mfcc':
                        self.npz_name = save_mfcc_feature(root=self.root, npz_name=self.npz_name, sample_rate=20e3,
                                                          signal_num=20e3, class_labels=TIDIGITS.classes)
                        self.data = load_mfcc_data(self.root, self.npz_name)
                        self.maxTime = 40
                        self.maxNum = int(self.data['neuron_num'])
                    else:
                        raise ValueError(">> Wrong preprocessing method. Please select kp or mfcc")

        else:
            raise ValueError(">> Failed to load the set, file not exist. You should download the dataset firstly.")

    def __getitem__(self, index):
        if self._is_train:
            if self.npz_name == 'kp_feature.npz':
                spiking = [self.data['train_audios'][index] * self.scale, self.data['train_ids'][index]]
            else:
                spiking = (self.data['train_audios'][index]).astype(float)
            label = np.int64(self.data['train_labels'][index])
        else:
            if self.npz_name == 'kp_feature.npz':
                spiking = [self.data['test_audios'][index] * self.scale, self.data['test_ids'][index]]
            else:
                spiking = (self.data['test_audios'][index]).astype(float)
            label = np.int64(self.data['test_labels'][index])
        return spiking, label

    def __len__(self):
        if self._is_train:
            return len(self.data['train_audios'])
        else:
            return len(self.data['test_audios'])

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def _dataset_exists(self):
        if os.path.exists(self.root):
            return True
        else:
            return False

    # 判断原始数据是否存在
    def _classfile_exists(self):
        for cls_id in TIDIGITS.classes.keys():
            if not os.path.isdir(os.path.join(self.root, 'test', cls_id)):
                return False
            if not os.path.isdir(os.path.join(self.root, 'train', cls_id)):
                return False
        return True

    def _npz_exists(self, npz_name):
        file_name = os.listdir(self.root)
        if npz_name == ('mfcc_feature', 'kp_feature'):
            if 'mfcc_feature.npz' in file_name:
                self.npz_name = 'mfcc_feature.npz'
            elif 'kp_feature.npz' in file_name:
                self.npz_name = 'kp_feature.npz'
            else:
                self.npz_name = ''
        else:
            if npz_name in file_name:
                self.npz_name = npz_name
            else:
                self.npz_name = ''




class SHD(Dataset):
    '''
    Spiking Heidelberg Digits Dataset
    max spiking time: 136.9 ms
    max neuron num: 700
    Class number: 20
    number of train samples: 8156
    number of test samples: 2264
    '''

    class_number = 20
    maxNum = 700
    Time = 1.37
    files = {
        "train_dataset": 'shd_train.h5',
        "test_dataset": 'shd_test.h5'
    }
    def __init__(self, root, is_train=True, **kwargs):
        super().__init__()
        self.scale = kwargs.get('scale', 100)
        self.maxTime = int(SHD.Time * self.scale)
        self.root = root
        self._is_train = is_train
        self.data = {
            'train_spiking': [],
            'test_spiking': [],
            'train_ids': [],
            'test_ids': [],
            'train_labels': [],
            'test_labels': []
        }
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")
        if self._dataset_exists():
            self._to_numpy_format()
        else:
            raise ValueError(">> Failed to load the set, file not exist. You should download the dataset firstly.")

    def __getitem__(self, index):
        if self._is_train:
            spiking = [self.data['train_spiking'][index] * self.scale, self.data['train_ids'][index]]
            label = np.int64(self.data['train_labels'][index])
        else:
            spiking = [self.data['test_spiking'][index] * self.scale, self.data['test_ids'][index]]
            label = np.int64(self.data['test_labels'][index])
        return spiking, label


    def __len__(self):
        if self._is_train:
            return len(self.data['train_spiking'])
        else:
            return len(self.data['test_spiking'])

    @property
    def dataset_folder(self):
        return os.path.join(self.root, self.__class__.__name__)

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def _to_numpy_format(self):
        # trainset
        # import tables
        import h5py
        if self._is_train:
            # train_fileh = tables.open_file(os.path.join(self.root, SHD.files['train_dataset']), mode='r')
            train_fileh = h5py.File(os.path.join(self.root, SHD.files['train_dataset']), 'r')
            neuron_ids = train_fileh['spikes']['units'][:]
            spike_times = train_fileh['spikes']['times'][:]
            labels = train_fileh['labels'][:]
            # neuron_ids = train_fileh.root.spikes.units
            # spike_times = train_fileh.root.spikes.times
            # labels = np.array(train_fileh.root.labels)

            self.data['train_spiking'] = spike_times
            self.data['train_ids'] = neuron_ids
            self.data['train_labels'] = labels

        else:
            # testset
            # test_fileh = tables.open_file(os.path.join(self.root, SHD.files['test_dataset']), mode='r')
            test_fileh = h5py.File(os.path.join(self.root, SHD.files['test_dataset']), 'r')
            neuron_ids = test_fileh['spikes']['units'][:]
            spike_times = test_fileh['spikes']['times'][:]
            labels = test_fileh['labels'][:]
            # neuron_ids = test_fileh.root.spikes.units
            # spike_times = test_fileh.root.spikes.times
            # labels = np.array(test_fileh.root.labels)

            self.data['test_spiking'] = spike_times
            self.data['test_ids'] = neuron_ids
            self.data['test_labels'] = labels

        print(">> Dataset loaded")

    def _dataset_exists(self):
        if os.path.exists(os.path.join(self.root)):
            for file in SHD.files.values():
                if not os.path.isfile(os.path.join(self.root, file)):
                    return False
            return True
        else:
            return False


class SSC(Dataset):
    '''
    Spiking Speech Command Dataset
    max spiking time: 99.95ms
    max neuron num: 700
    Class number: 35
    number of train samples: 75466
    number of test samples: 30363
    '''

    class_number = 35
    maxNum = 700
    Time = 1
    files = {
        "train_dataset": 'ssc_train.h5',
        "valid_dataset": 'ssc_valid.h5',
        "test_dataset": 'ssc_test.h5'
    }

    def __init__(self, root, is_train=True, **kwargs):
        super().__init__()
        self.scale = kwargs.get('scale', 100)
        self.maxTime = int(SSC.Time * self.scale)
        self.root = root
        self._is_train = is_train
        self.data = {
            'train_spiking': [],
            'test_spiking': [],
            'train_ids': [],
            'test_ids': [],
            'train_labels': [],
            'test_labels': []
        }
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")
        if self._dataset_exists():
            self._to_numpy_format()
        else:
            raise ValueError(">> Failed to load the set, file not exist. You should download the dataset firstly.")


    def __getitem__(self, index):
        if self._is_train:
            spiking = [self.data['train_spiking'][index] * self.scale, self.data['train_ids'][index]]
            label = np.int64(self.data['train_labels'][index])
        else:
            spiking = [self.data['test_spiking'][index] * self.scale, self.data['test_ids'][index]]
            label = np.int64(self.data['test_labels'][index])
        return spiking, label

    def __len__(self):
        if self._is_train:
            return len(self.data['train_spiking'])
        else:
            return len(self.data['test_spiking'])

    @property
    def dataset_folder(self):
        return os.path.join(self.root, self.__class__.__name__)

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def _to_numpy_format(self):
        # import tables
        import h5py

        # trainset
        if self._is_train:
            # train_fileh = tables.open_file(os.path.join(self.root, SSC.files['train_dataset']), mode='r')
            train_fileh = h5py.File(os.path.join(self.root, SSC.files['train_dataset']), 'r')
            neuron_ids = train_fileh['spikes']['units'][:]
            spike_times = train_fileh['spikes']['times'][:]
            labels = train_fileh['labels'][:]
            # neuron_ids = train_fileh.root.spikes.units
            # spike_times = train_fileh.root.spikes.times
            # labels = np.array(train_fileh.root.labels)

            self.data['train_spiking'] = spike_times
            self.data['train_ids'] = neuron_ids
            self.data['train_labels'] = labels

        else:
            # test_fileh = tables.open_file(os.path.join(self.root, SSC.files['test_dataset']), mode='r')
            test_fileh = h5py.File(os.path.join(self.root, SSC.files['test_dataset']), 'r')
            neuron_ids = test_fileh['spikes']['units'][:]
            spike_times = test_fileh['spikes']['times'][:]
            labels = test_fileh['labels'][:]
            # neuron_ids = test_fileh.root.spikes.units
            # spike_times = test_fileh.root.spikes.times
            # labels = np.array(test_fileh.root.labels)

            self.data['test_spiking'] = spike_times
            self.data['test_ids'] = neuron_ids
            self.data['test_labels'] = labels

        print(">> Dataset loaded")

    def _dataset_exists(self):
        if os.path.exists(os.path.join(self.root)):
            for file in SSC.files.values():
                if not os.path.isfile(os.path.join(self.root, file)):
                    return False
            return True
        else:
            return False

class DVS128Gesture(Dataset):
    """
    参考 https://spikingjelly.readthedocs.io/zh_CN/latest/clock_driven/13_neuromorphic_datasets.html
    """
    resources = 'https://ibm.ent.box.com/s/3hiq58ww1pbbjrinh367ykfdf60xsfm8/folder/50167556794'
    scale = 0.001
    resize = 1
    class_number = 11
    maxNum = (128 // resize) * (128 // resize)
    # maxTime: 18456.951ms 数据集中单位为us
    origin_maxTime = 18457
    maxTime = 1000  # mstb编码也设置为最长1秒
    maxFrameValue = 65.0
    files = {
        "train_dataset": 'trials_to_train.txt',
        "test_dataset": 'trials_to_test.txt'
    }

    def __init__(self, dataset_root, is_train=True, data_type='event', event_cut=None, frames_number=None,
                 split_by=None, duration=None, time_step=None):
        """
        Args:
            dataset_root:
            is_train:
            data_type:
            event_cut: 事件太长需要切分，分割事件的时间长度，ms
            frames_number:
            split_by:
            duration:
        """
        super().__init__()
        self.root = dataset_root
        self._is_train = is_train
        self.data_type = data_type
        self.frame_number = frames_number
        self.duration = duration
        self.time_step = time_step
        self.data = {
            'train_spiking': [],
            'test_spiking': [],
            'train_ids': [],
            'test_ids': [],
            'train_labels': [],
            'test_labels': []
        }
        self.file_list = {
            'train_files': [],
            'test_files': []
        }
        if event_cut is None:
            events_np_root = os.path.join(dataset_root, 'events_np')
        else:
            events_np_root = os.path.join(dataset_root, f'events_np_{event_cut}')
        self.events_np_root = events_np_root
        if not isinstance(self._is_train, bool):
            raise TypeError(">> is_train should be boolean value")
        train_path = os.path.join(events_np_root, 'train')
        test_path = os.path.join(events_np_root, 'test')
        if self._dataset_exists():
            if os.path.exists(train_path) and os.path.exists(test_path):
                print("DVS128 Gesture Dataset had been transferred into npz format before.")
            else:
                os.mkdir(events_np_root)
                print(f'Mkdir [{events_np_root}].')
                self.create_events_np_files(events_np_root, event_cut)
        else:
            raise ValueError(">> Faild to load the set, file not exist. You should download the dataset firstly.")
        self.H, self.W = self.get_H_W()
        # self._to_numpy_format()

        if data_type == 'event':
            for root, dirs, files in os.walk(train_path):
                for file in files:
                    self.file_list['train_files'].append(os.path.join(root, file))
            for root, dirs, files in os.walk(test_path):
                for file in files:
                    self.file_list['test_files'].append(os.path.join(root, file))
        elif data_type == 'frame':
            if frames_number is not None:
                assert frames_number > 0 and isinstance(frames_number, int)
                assert split_by == 'time' or split_by == 'number'
                frames_np_root = os.path.join(dataset_root, f'frames_number_{frames_number}_split_by_{split_by}')
                if os.path.exists(frames_np_root):
                    print(f'The directory [{frames_np_root}] already exists.')
                else:
                    os.mkdir(frames_np_root)
                    print(f'Mkdir [{frames_np_root}].')
                    # create the same directory structure
                    create_same_directory_structure(events_np_root, frames_np_root)

                    # use multi-thread to accelerate
                    t_ckp = time.time()
                    with ThreadPoolExecutor(max_workers=min(multiprocessing.cpu_count(), 64)) as tpe:
                        print(f'Start ThreadPoolExecutor with max workers = [{tpe._max_workers}].')
                        for e_root, e_dirs, e_files in os.walk(events_np_root):
                            if e_files.__len__() > 0:
                                output_dir = os.path.join(frames_np_root, os.path.relpath(e_root, events_np_root))
                                for e_file in e_files:
                                    events_np_file = os.path.join(e_root, e_file)
                                    print(
                                        f'Start to integrate [{events_np_file}] to frames and save to [{output_dir}].')
                                    tpe.submit(integrate_events_file_to_frames_file_by_fixed_frames_number,
                                               events_np_file, output_dir, split_by, frames_number, self.H, self.W,
                                               True)

                    print(f'Used time = [{round(time.time() - t_ckp, 2)}s].')

                train_path = os.path.join(frames_np_root, 'train')
                test_path = os.path.join(frames_np_root, 'test')
                for root, dirs, files in os.walk(train_path):
                    for file in files:
                        self.file_list['train_files'].append(os.path.join(root, file))
                for root, dirs, files in os.walk(test_path):
                    for file in files:
                        self.file_list['test_files'].append(os.path.join(root, file))

            elif duration is not None:
                # duration单位为us
                assert duration > 0 and isinstance(duration, int)
                frames_np_root = os.path.join(dataset_root, f'duration_{duration}')
                if os.path.exists(frames_np_root):
                    print(f'The directory [{frames_np_root}] already exists.')

                else:
                    os.mkdir(frames_np_root)
                    print(f'Mkdir [{frames_np_root}].')
                    # create the same directory structure
                    create_same_directory_structure(events_np_root, frames_np_root)
                    # use multi-thread to accelerate
                    t_ckp = time.time()
                    with ThreadPoolExecutor(max_workers=min(multiprocessing.cpu_count(), 64)) as tpe:
                        print(f'Start ThreadPoolExecutor with max workers = [{tpe._max_workers}].')
                        for e_root, e_dirs, e_files in os.walk(events_np_root):
                            if e_files.__len__() > 0:
                                output_dir = os.path.join(frames_np_root, os.path.relpath(e_root, events_np_root))
                                for e_file in e_files:
                                    events_np_file = os.path.join(e_root, e_file)
                                    print(
                                        f'Start to integrate [{events_np_file}] to frames and save to [{output_dir}].')
                                    tpe.submit(integrate_events_file_to_frames_file_by_fixed_duration, events_np_file,
                                               output_dir, duration, self.H, self.W, True)

                    print(f'Used time = [{round(time.time() - t_ckp, 2)}s].')

                train_path = os.path.join(frames_np_root, 'train')
                test_path = os.path.join(frames_np_root, 'test')
                for root, dirs, files in os.walk(train_path):
                    for file in files:
                        self.file_list['train_files'].append(os.path.join(root, file))
                for root, dirs, files in os.walk(test_path):
                    for file in files:
                        self.file_list['test_files'].append(os.path.join(root, file))
            else:
                raise ValueError('frames_number and duration can not both be None.')

    def calculate_max_time(self):
        """
        读取数据集，计算事件的最大持续时间
        Returns:
        """
        if os.path.exists(os.path.join(self.root)):
            data_files = {
                'train_dataset': [],
                'test_dataset': []
            }
            with open(os.path.join(self.root, self.files['train_dataset'])) as train_f:
                for line in train_f.readlines():
                    line = line.strip('\n')
                    data_files['train_dataset'].append(line.split('.')[0] + '_labels.csv')
            with open(os.path.join(self.root, self.files['test_dataset'])) as test_f:
                for line in test_f.readlines():
                    line = line.strip('\n')
                    data_files['test_dataset'].append(line.split('.')[0] + '_labels.csv')
            for files in data_files.values():
                for f in files:
                    fpath = os.path.join(self.root, f)
                    with open(fpath, 'r') as cf:
                        import csv
                        csv_reader = csv.reader(cf)
                        birth_header = next(csv_reader)  # 读取第一行每一列的标题
                        for row in csv_reader:  # 将csv 文件中的数据保存到birth_data中
                            self.maxTime = max(self.maxTime, int(row[2]) - int(row[1]))
            print('maxTime:', self.maxTime / 1000)  # maxTime: 18456.951
        else:
            raise ValueError(">> Faild to load the set, file not exist. You should download the dataset firstly.")

    def calculate_max_value_from_frame(self):
        train_path = self.file_list['train_files']
        test_path = self.file_list['test_files']
        maxFrameValue = 0
        for f in train_path:
            data = np.load(f)['frames']
            maxFrameValue = max(maxFrameValue, np.max(data))
        for f in test_path:
            data = np.load(f)['frames']
            maxFrameValue = max(maxFrameValue, np.max(data))
        print('maxFrameValue:', maxFrameValue)  # maxFrameValue: 53.0

    def calculate_max_time_from_npz(self):
        """
        读取数据集，计算事件的最大持续时间
        Returns:
        """
        if os.path.exists(os.path.join(self.events_np_root)):
            train_path = os.path.join(self.events_np_root, 'train')
            test_path = os.path.join(self.events_np_root, 'test')
            for root, dirs, files in os.walk(train_path):
                for file in files:
                    data = np.load(os.path.join(root, file))
                    t = data['t']
                    self.maxTime = max(self.maxTime, np.max(t) - np.min(t))
            for root, dirs, files in os.walk(test_path):
                for file in files:
                    data = np.load(os.path.join(root, file))
                    t = data['t']
                    self.maxTime = max(self.maxTime, np.max(t) - np.min(t))
            print('maxTime:', self.maxTime / 1000)  # maxTime: 18456.951 ms
        else:
            raise ValueError(">> Faild to load the set, file not exist. You should download the dataset firstly.")

    def __getitem__(self, index):
        if self._is_train:
            file_name = self.file_list['train_files'][index]

        else:
            file_name = self.file_list['test_files'][index]

        data = np.load(file_name)

        if self.data_type == 'event':
            label = int(os.path.split(file_name)[-1].split('.')[0].split('_')[-1])
            t = data['t']
            x = data['x'] // self.resize
            y = data['y'] // self.resize
            p = data['p']
            index_p = (p == 1)
            x = x[index_p]
            y = y[index_p]
            t = t[index_p]
            p = p[index_p]
            neuron_index = []
            firing_time = []
            new_w = self.W // self.resize
            for i in range(len(x)):
                if t[i] - t[0] > 1e6:  # 只取一秒
                    break
                neuron_index.append(y[i] * new_w + x[i])
                firing_time.append(t[i] - t[0])

            spiking = [np.asarray(firing_time) * DVS128Gesture.scale, neuron_index]
            # label = np.int64(self.data['train_labels'][index])
        elif self.data_type == 'frame' and self.duration is not None:
            import random
            frames = data['frames']
            # 随机取一段数据
            nums = frames.shape[0]
            left = random.randint(0, nums - self.time_step)
            frames = frames[left: left + self.time_step, ...]
            label = int(os.path.split(file_name)[-1].split('.')[0].split('_')[-2])
            # normalization
            # spiking = frames / self.maxFrameValue
            spiking = (frames > 0) * 1
            spiking = spiking.astype(np.float32)  # shape[batch, time_step, channel, H ,W], 在编码中交换维度
        elif self.data_type == 'frame' and self.frame_number is not None:
            import random
            spiking = data['frames']
            spiking = spiking.astype(np.float32)
            label = int(os.path.split(file_name)[-1].split('.')[0].split('_')[-1])
        else:
            raise NotImplementedError
        return spiking, label

    def __len__(self):
        if self._is_train:
            return len(self.file_list['train_files'])
        else:
            return len(self.file_list['test_files'])

    def split_event_to_short(self, spiking):
        bins = spiking.shape[-1] // self.maxTime

    @property
    def dataset_folder(self):
        return os.path.join(self.root, self.__class__.__name__)

    @property
    def data_dict(self):
        return self.data

    @property
    def is_train(self):
        return self._is_train

    @is_train.setter
    def is_train(self, is_train):
        assert is_train in [True, False], ">> Invalid is_train setting"
        self._is_train = is_train

    def _dataset_exists(self):
        if os.path.exists(os.path.join(self.root)):
            data_files = {
                'train_dataset': [],
                'test_dataset': []
            }
            with open(os.path.join(self.root, self.files['train_dataset'])) as train_f:
                for line in train_f.readlines():
                    line = line.strip('\n')
                    data_files['train_dataset'].append(line)
            with open(os.path.join(self.root, self.files['test_dataset'])) as test_f:
                for line in test_f.readlines():
                    line = line.strip('\n')
                    data_files['test_dataset'].append(line)
            for files in data_files.values():
                for f in files:
                    if not os.path.isfile(os.path.join(self.root, f)):
                        return False
            return True
        else:
            return False

    @staticmethod
    def split_aedat_files_to_np(fname: str, aedat_file: str, csv_file: str, output_dir: str, event_split=None):
        events = load_aedat_v3(aedat_file)
        print(f'Start to split [{aedat_file}] to samples.')
        # Read csv file and get time stamp and label of each sample. Then split the origin data to samples
        csv_data = np.loadtxt(csv_file, dtype=np.uint32, delimiter=',', skiprows=1)

        # Note that there are some files that many samples have the same label, e.g., user26_fluorescent_labels.csv
        label_file_num = [0] * 11
        if event_split is not None:
            event_split *= 1000  # convert ms to us
        for i in range(csv_data.shape[0]):
            # the label of DVS128 Gesture is 1, 2, ..., 11. We set 0 as the first label, rather than 1
            label = csv_data[i][0] - 1
            t_start = csv_data[i][1]
            t_end = csv_data[i][2]
            mask = np.logical_and(events['t'] >= t_start, events['t'] < t_end)
            if event_split is None:
                file_name = os.path.join(output_dir, str(label), f'{fname}_{label}.npz')
                np.savez(file_name,
                         t=events['t'][mask],
                         x=events['x'][mask],
                         y=events['y'][mask],
                         p=events['p'][mask]
                         )
                print(f'[{file_name}] saved.')
                label_file_num[label] += 1
            else:
                total_time = t_end - t_start
                bins = total_time // event_split
                t = events['t'][mask]
                x = events['x'][mask]
                y = events['y'][mask]
                p = events['p'][mask]
                left = 0
                if bins == 0:
                    bins = 1  # 事件长度小于指定的时间
                per_bin = t.shape[0] // bins
                for j in range(bins):
                    right = left + per_bin
                    file_name = os.path.join(output_dir, str(label), f'{fname}_{j}_{label}.npz')
                    np.savez(file_name,
                             t=t[left: right],
                             x=x[left: right],
                             y=y[left: right],
                             p=p[left: right],
                             )
                    print(f'[{file_name}] saved.')
                    left = right
                    label_file_num[label] += 1

    def create_events_np_files(self, events_np_root: str, event_split=None):
        '''
        :param events_np_root: Root directory path which saves events files in the ``npz`` format
        :type events_np_root: str
        This function defines how to convert the origin binary data in ``extract_root`` to ``npz`` format and save converted files in ``events_np_root``.
        Args:
            event_split: 事件划分成几份
        '''

        train_dir = os.path.join(events_np_root, 'train')
        test_dir = os.path.join(events_np_root, 'test')
        os.mkdir(train_dir)
        os.mkdir(test_dir)
        print(f'Mkdir {train_dir, test_dir}.')
        for label in range(11):
            os.mkdir(os.path.join(train_dir, str(label)))
            os.mkdir(os.path.join(test_dir, str(label)))
        print(f'Mkdir {os.listdir(train_dir)} in [{train_dir}] and {os.listdir(test_dir)} in [{test_dir}].')

        with open(os.path.join(self.root, DVS128Gesture.files['train_dataset'])) as trials_to_train_txt, open(
                os.path.join(self.root, DVS128Gesture.files['test_dataset'])) as trials_to_test_txt:
            # use multi-thread to accelerate
            # t_ckp = time.time()
            with ThreadPoolExecutor(max_workers=min(multiprocessing.cpu_count(), 64)) as tpe:
                print(f'Start the ThreadPoolExecutor with max workers = [{tpe._max_workers}].')

                for fname in trials_to_train_txt.readlines():
                    fname = fname.strip()
                    if fname.__len__() > 0:
                        aedat_file = os.path.join(self.root, fname)
                        fname = os.path.splitext(fname)[0]
                        tpe.submit(DVS128Gesture.split_aedat_files_to_np, fname, aedat_file,
                                   os.path.join(self.root, fname + '_labels.csv'), train_dir)

                for fname in trials_to_test_txt.readlines():
                    fname = fname.strip()
                    if fname.__len__() > 0:
                        aedat_file = os.path.join(self.root, fname)
                        fname = os.path.splitext(fname)[0]
                        tpe.submit(DVS128Gesture.split_aedat_files_to_np, fname, aedat_file,
                                   os.path.join(self.root, fname + '_labels.csv'), test_dir)

            # print(f'Used time = [{round(time.time() - t_ckp, 2)}s].')
        print(f'All aedat files have been split to samples and saved into [{train_dir, test_dir}].')

    def get_H_W(self):
        '''
        :return: A tuple ``(H, W)``, where ``H`` is the height of the data and ``W` is the weight of the data.
            For example, this function returns ``(128, 128)`` for the DVS128 Gesture dataset.
        :rtype: tuple
        '''
        return 128, 128




# encoding: utf-8
"""
@author: Yuan Mengwen
@contact: mwyuan94@gmail.com
@project: PyCharm
@filename: Decoders.py
@time:2021/5/7 14:50
@description:
"""

from .Node import Node, Decoder
import torch
import numpy as np


class Spike_Counts(Decoder):

    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Spike_Counts, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.pop_size = kwargs.get('pop_size', 1)
        self.bias = kwargs.get('bias', 0.0)
        self.scale = kwargs.get('scale', 1.0)

    def numpy_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        spike_rate = record.mean(0)
        pop_num = int(self.num / self.pop_size)
        pop_spikes_temp = (
            [
                np.sum(spike_rate[:, (i * self.pop_size): (i * self.pop_size) + self.pop_size], axis=1)
                for i in range(pop_num)
            ]
        )
        pop_spikes = np.stack(pop_spikes_temp, axis=1)
        # spikes_list = pop_spikes.tolist()
        # max_value = np.max(pop_spikes, 1)
        # batch_size = pop_spikes.shape[0]
        # predict_labels = []
        # for i in range(batch_size):
        #     index = spikes_list[i].index(max_value[i])
        #     predict_labels.append(index)
        return pop_spikes

    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        if '[2]' in self.coding_var_name:
            pop_spikes = record[:,:,0,:].sum(0).to(device=device)
        else:
            spike_rate = record.sum(0).to(device=device)
            pop_num = int(self.num / self.pop_size)
            pop_spikes_temp = (
                [
                    spike_rate[:, (i * self.pop_size): (i * self.pop_size) + self.pop_size].sum(dim=1)
                    for i in range(pop_num)
                ]
            )
            pop_spikes = torch.stack(pop_spikes_temp, dim=-1)
        return (pop_spikes + self.bias)*self.scale

Decoder.register('spike_counts', Spike_Counts)

class Final_Step_Voltage(Decoder):

    """
    Get label that has the highest voltage.
    """

    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='V', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Final_Step_Voltage, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        final_step_voltage = record[-1, :, :]

        return final_step_voltage

Decoder.register('final_step_voltage', Final_Step_Voltage)

class First_Spike(Decoder):


    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(First_Spike, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    # def numpy_decoding(self, record):
    #     # the shape of record is (time_step, batch_size, n_neurons)
    #     # get predict label
    #     spikes_list = record.tolist()
    #
    #     return predict_matrix

    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        record_temp = record.permute(1, 0, 2)
        [batch_size, time_step, n_neurons] = record_temp.shape
        batch_index = []
        for i in range(batch_size):
            index = torch.nonzero(record_temp[i, ::])
            if len(index) == 0:
                first_spike_row = 0 + i * time_step
            else:
                first_spike_row = index[0, 0].item() + i * time_step

            batch_index.append(first_spike_row)
        batch_index = torch.LongTensor(batch_index).to(device=device)
        record_temp = record_temp.reshape(batch_size*time_step, n_neurons)
        predict_matrix = torch.index_select(record_temp, 0, batch_index).to(device=device)
        return predict_matrix

Decoder.register('first_spike', First_Spike)

class TimeSpike_Counts(Decoder):


    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(TimeSpike_Counts, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def numpy_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        # get predict label
        record = np.sum(record, axis=-1)
        spikes_list = record.tolist()
        max_value = np.max(record, 1)
        batch_size = record.shape[0]
        predict_labels = []
        for i in range(batch_size):
            index = spikes_list[i].index(max_value[i])
            predict_labels.append(index)
        return predict_labels

    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        tlen = record.shape[0]
        tt = torch.arange(0, tlen, device=device, dtype=torch.float)
        tw = 0.1*torch.exp(-tt/(0.5*tlen))
        predict_labels = torch.sum(record.permute(1, 2, 0)*tw, dim=-1)
        return predict_labels

Decoder.register('time_spike_counts',TimeSpike_Counts)

class NullDeocder(Decoder):
    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(NullDeocder, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def torch_coding(self, record, target, device):
        return record
Decoder.register('null', NullDeocder)


class V_Trajectory(Decoder):

    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='V', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(V_Trajectory, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)

        return record

Decoder.register('v_t', V_Trajectory)


class Time_Softmax(Decoder):


    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Time_Softmax, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def numpy_coding(self, record, target, device):
        pass
      
    def torch_coding1(self, record: torch.Tensor, target, device):
        def grad_regulate_hook(grad):
            return grad - torch.mean(grad)

        spike_i = record[:,:,0,...]
        spike_t = record[:,:,1,...]
        old_shape = spike_t.shape
        tlen = record.shape[0]
        time_array = self.dt*torch.arange(0, tlen, device=device, dtype=torch.float)
        spike_t = spike_i.detach()*time_array.view(-1, 1, 1) - spike_t# + (1-spike_i.detach())*1000.0
        max_t, ind = torch.max(spike_t.permute(0, 2, 1).reshape(-1, old_shape[1]).detach(), dim=0)
        mshape = [1, -1] + [1]*(spike_t.dim()-2)
        spike_t = 0.1*(max_t.view(mshape) + (spike_i - spike_i.detach()) - spike_t)*spike_i.detach()
        # spike_t.register_hook(grad_regulate_hook)
        # spike_t[:,0] = 7
        out = torch.softmax(torch.norm(spike_t, dim=0), dim=-1)
        # torch.sum(spike_t, dim=-1)     *spike_t.view(old_shape[0],-1)
        # from matplotlib import pyplot as plt
        # plt.plot(out.detach().cpu().numpy()[0,:])
        # plt.show()
        # out = out.view(old_shape)
        # out = torch.sum(out, dim=-1)
        # if torch.mean(out) <1:
        return out

    def torch_coding(self, record: torch.Tensor, target, device):
        # def grad_regulate_hook(grad):
        #     return grad - torch.mean(grad)

        shape = list(record.shape)
        shape[-1] = self.num
        shape.append(-1)
        record = record.view(*shape)


        spike_i = record[:,:,0,...]
        spike_t = record[:,:,1,...]
        spike = spike_t + spike_i
        # spk_ind = spike_t.clamp_max(1.0).detach()
        # old_shape = spike_t.shape
        tlen = record.shape[0]
        time_array = self.dt*torch.arange(0, tlen, device=device, dtype=torch.float)
        spike_t = time_array.view(-1, 1, 1, 1) - spike + spike.detach() + spike_i.le(0.0)*self.dt*tlen
        # spike_t = spike_i.gt(0.0)*time_array.view(-1, 1, 1) - spike_t
        frist_times = torch.amin(spike_t, dim=(0,2), keepdim=True).detach()
        # spike_ft = torch.amin(spike_t, dim=0, keepdim=True)
        # target_frist = torch.gather(frist_times, 2, target.view(1,-1,1,1)).detach()
        out_count = torch.mean(torch.sum(spike_i, dim=0), dim=-1).detach() + 1.0
        out_ti = 5.0*torch.exp((frist_times-spike_t)/50.0)
        out = torch.mean(torch.sum(out_ti,dim=0),dim=-1)
        out = out/out_count + out.detach()*(out_count-1)/out_count
        rateloss = torch.norm(torch.mean(torch.sum(spike_i,dim=0), dim=-2)-3.0)
        return out, rateloss

    def torch_coding3(self, record: torch.Tensor, target, device):
        # def grad_regulate_hook(grad):
        #     return grad - 10000#torch.mean(grad)
        spike_i = record[:,:,0,...]
        spike_t = record[:,:,1,...]
        self.out = torch.sum(record, dim=2)
        self.target = target.repeat(spike_t.shape[0], 1).unsqueeze(-1)


        with torch.no_grad():
            tlen = record.shape[0]
            time_array = self.dt*torch.arange(0, tlen, device=device, dtype=torch.float)
            spike_t = time_array.view(-1, 1, 1) - spike_t
            exp_n = spike_i*torch.exp(-spike_t/20.0)
            exp_sum = torch.sum(exp_n, dim=(0, 2), keepdim=True) + 1.0e-20
            self.exp_n = exp_n/exp_sum

        return torch.sum(self.exp_n, dim=0)

    @property
    def loss(self):
        with torch.no_grad():
            pos_n = torch.gather(self.exp_n, dim=2, index=self.target)
            sum_pos = torch.sum(pos_n, dim=0, keepdim=True) + 1.0e-20
            pos_n = (sum_pos - 1) * (pos_n + 1.0e-18) / sum_pos
            out_grad = torch.scatter(self.exp_n * torch.gt(sum_pos, 1.0e-10), 2, self.target, pos_n)
        if self.out.requires_grad == True:
            self.out.backward(out_grad)
        return  -torch.mean(torch.log(sum_pos))

Decoder.register('time_softmax', Time_Softmax)

class Voltage_Sum(Decoder):

    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='V', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Voltage_Sum, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)

    def numpy_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        record_temp = record.sum(0)
        spikes_list = record_temp.tolist()
        max_value = np.max(record_temp, 1)
        batch_size = record_temp.shape[0]
        predict_labels = []
        for i in range(batch_size):
            index = spikes_list[:, i].index(max_value[i])
            predict_labels.append(index)
        return predict_labels

    def torch_coding(self, record, target, device):
        # the shape of record is (time_step, batch_size, n_neurons)
        predict_matrix = record[-1,...].to(device=device)
        mp = torch.mean(predict_matrix).detach()
        out = (predict_matrix - mp)
        # predict_matrix = predict_temp.permute(1, 0)
        return out

Decoder.register('voltage_sum', Voltage_Sum)

class Complex_Count(Decoder):
    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Complex_Count, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.tlen = None
    def torch_coding(self, record: torch.Tensor, target, device):
        assert record.dtype.is_complex
        if self.tlen is None:
            self.tlen = record.shape[0]*1.0
        time_array = torch.arange(0, self.tlen, device=device, dtype=torch.float).view(-1,1,1)
        spk = record.real.gt(0.0)
        # for gradient test
        sum_spk = torch.cumsum(spk, dim=0)
        spk = spk*sum_spk.lt(5)

        # count = record.imag.gt(0)
        out = torch.sum(spk*(-(time_array-record.imag)/self.tlen), dim=0) + 1.0e-6
        # out = torch.sum(count, dim=0) + 1.0e-6
        # rate = torch.sum(count, dim=0)/self._backend.time
        return out

Decoder.register('complex_count', Complex_Count)


class Complex_Phase(Decoder):
    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Complex_Phase, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.trange = kwargs.get('trange',2.0)
        self.period = kwargs.get('period', 20.0)
        # class TransClamp(torch.autograd.Function):
        #     @staticmethod
        #     def forward(ctx, x, min=None, max=None):
        #         return torch.clamp(x, min, max)
        #
        #     @staticmethod
        #     def backward(ctx, grad_outputs):
        #         return grad_outputs, None, None
        # self.clamp = TransClamp.apply




    def torch_coding(self,  record: torch.Tensor, target, device):
        assert record.dtype.is_complex
        # record.shape = (time, batch, num)
        tlen = record.shape[0]
        reference = torch.mean(record.abs(), dim=-1, keepdim=True)
        reference = torch.softmax(reference, dim=0)*self.dt
        phase = record.real/(record.abs()+1.0)
        phase = torch.sum(phase*reference, dim=0)
        return phase

    def torch_coding2(self, record, target, device):
        import torch.nn.functional as F
        tlen = record.shape[0]
        batch_size = record.shape[1]
        out_num = record.shape[2]
        kernel_range = int(self.period / self.dt)
        w = 2 * np.pi / self.period
        tt = self.dt * torch.arange(0, kernel_range, device=device, dtype=torch.float64)
        kernel_i = torch.sin(-w * tt).view(1, 1, -1)
        kernel_t = -self.dt*w*torch.cos(-w * tt).view(1, 1, -1)

        x = record.view(tlen, -1).t().view(-1, 1, tlen)
        conv_i = torch.mean(F.conv1d(x.real, kernel_i, padding=kernel_range)[...,kernel_range//2:kernel_range//2+tlen]
                            .view(batch_size, out_num, tlen), dim=1, keepdim=True)
        conv_t = torch.mean(F.conv1d(x.real, kernel_t, padding=kernel_range)[...,kernel_range//2:kernel_range//2+tlen]
                            .view(batch_size, out_num, tlen), dim=1, keepdim=True)
        x = x.view(batch_size, out_num, tlen)
        out_phase = torch.sum(conv_i*x.real.detach() + conv_t*x.imag*self.dt, dim=-1)

        return out_phase, conv_i


Decoder.register('complex_phase', Complex_Phase)

class Complex_Latency(Decoder):
    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Complex_Latency, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.tlen = None

    def torch_coding(self,  record: torch.Tensor, target, device):
        assert record.dtype.is_complex
        # if self.tlen is None:
        self.tlen = record.shape[0]
        time_array = torch.arange(0, self.tlen, device=device, dtype=torch.float).view(-1,1,1)
        spk = record.real.gt(0)
        spk_time = (self.tlen-(time_array - record.imag))*spk
        spk_time = torch.exp(1*(spk_time-torch.amax(spk_time, dim=(0,2), keepdim=True).detach())/self.tlen)
        spk_rate = record.real
        spk_weight = torch.exp(-torch.cumsum(record.real.detach(),dim=0)/5.0)
        weighted_spk_time = torch.sum(spk_weight*spk_rate*spk_time*spk,dim=0)
        return weighted_spk_time

Decoder.register('complex_latency', Complex_Latency)

class Complex_TimingDistance(Decoder):
    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Complex_TimingDistance, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        from matplotlib import pyplot as plt
        self.tlen = None
        self.filter_time = kwargs.get("filter_time", 10.0)


    def build(self, backend):
        super(Complex_TimingDistance, self).build(backend)
        tdt = 6.0*self.dt/(1.0*self.filter_time)
        tt = torch.arange(0, 6.0+tdt, tdt)
        self.rate_filter = (torch.exp(-(tt - 2) ** 2)).view(1, 1, -1)
        self.d_rate_filter = (0.5*(tt-2)*torch.exp(-(tt - 1) ** 2)).view(1, 1, -1)*tdt*0.01


    def torch_coding(self, record: torch.Tensor, target: torch.Tensor , device: str):
        from torch.nn.functional import mse_loss
        # record shape (time, batch, neuron)
        # target shape (batch, neuron, time)
        assert record.dtype.is_complex
        n_time, n_batch, n_neuron = record.shape
        self.rate_filter = self.rate_filter.to(device)
        self.d_rate_filter = self.d_rate_filter.to(device)
        target = target.view(-1, 1, n_time)
        record = record.permute(1,2,0).view(-1, 1, n_time)

        target_rate = torch.conv1d(target, self.rate_filter, padding='same')
        record_rate = torch.conv1d(record.real, self.rate_filter, padding='same') - torch.conv1d(record.imag, self.d_rate_filter, padding='same')
        rate_loss = mse_loss(record_rate, target_rate)
        return rate_loss, record_rate, target_rate

Decoder.register('complex_timing_distance', Complex_TimingDistance)




class Complex_Trajectory(Decoder):

    def __init__(self, num=None, dec_target=None, dt=None, coding_method=('poisson', 'spike_counts', '...'), coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), **kwargs):
        super(Complex_Trajectory, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.tau_d = kwargs.get('tau_d', 20.0+20*torch.rand(num).view(1,-1))
        self.tau_r = kwargs.get('tau_r', 10.0+50*torch.rand(num).view(1,-1))
        self.group_num = kwargs.get('group_num', 1)
        self.num = num
        assert num%self.group_num == 0
        if not isinstance(self.tau_d, torch.Tensor):
            self.tau_d = torch.tensor(self.tau_d)
        if not isinstance(self.tau_r, torch.Tensor):
            self.tau_r = torch.tensor(self.tau_r)
        self.weight = kwargs.get('weight', None)
        if self.weight is None:
            self.weight = torch.randn(num).view(1,-1)

    def build(self, backend):
        super(Complex_Trajectory, self).build(backend)
        tau_d = self.variable_to_backend(self.id+'_Complex_Trajectory_tau_d', self.tau_d.shape, self.tau_d, True)
        tau_r = self.variable_to_backend(self.id+'_Complex_Trajectory_tau_r', self.tau_r.shape, self.tau_r, True)
        weight = self.variable_to_backend(self.id+'_Complex_Trajectory_weight', self.tau_r.shape, self.weight, True)
        self.tau_d = tau_d.value
        self.tau_r = tau_r.value
        self.weight = weight.value




    def torch_coding(self,  record: torch.Tensor, target=None, device='cpu'):
        decay = torch.exp(-self.dt/self.tau_d).to(device)
        rota = (2.0*torch.pi*self.dt/self.tau_r).to(device)
        complex_beta = torch.view_as_complex(torch.stack([decay * torch.cos(-rota),
                                                          decay * torch.sin(-rota)], dim=-1))
        weight = self.weight.unsqueeze(-1)
        tlen = record.shape[0]
        Xs = []
        x = torch.zeros_like(record[0])
        for ii in range(tlen):
            x = complex_beta*x + record[ii].real*(0.0+1.0j)*complex_beta**record[ii].imag
            Xs.append(x)
        Xs = weight*torch.stack(Xs, dim=-1)
        Xs = Xs.view(-1, self.group_num, self.num//self.group_num, tlen)

        # trace = Xs.real
        # dtrace = rota*Xs.imag - Xs.real/self.tau_d

        return torch.sum(Xs.real, dim=-2)

Decoder.register('complex_trajectory', Complex_Trajectory)


class Complex_Spike_Conv(Decoder):
    def __init__(self, num=None, dec_target=None, dt=None, coding_method='spike_counts', coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), decay=0.9, ocillate=-0.01, **kwargs):
        super(Complex_Spike_Conv, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.kernel = kwargs.get('kernel', None)
        self.time_window = kwargs.get('time_window', 80)/2.0

        if self.kernel is None:
            tt = torch.arange(-self.time_window, self.time_window, self.dt)
            self.kernel = torch.exp(-(3*tt/self.time_window)**2).view(1, 1, -1)
            self.d_kernel = (6*tt/self.time_window)*torch.exp(-(3*tt/self.time_window)**2).view(1, 1, -1)
        self.klen = self.kernel.shape[-1]

    def torch_coding(self, source, target, device='cpu'):
        weight = self.kernel.to(device).expand(source.shape[1],source.shape[1],self.klen)
        d_weight = self.d_kernel.to(device).expand(source.shape[1],source.shape[1],self.klen)
        source = source.transpose(0, 1).transpose(1, 2)
        target = target.transpose(0, 1).transpose(1, 2)
        filt_source = torch.conv1d(source.real.to(weight.dtype), weight, padding='same') + torch.conv1d((source.real*source.imag).to(weight.dtype)*self.dt, d_weight, padding='same')
        filt_target = torch.conv1d(target.real.to(weight.dtype), weight, padding='same') + torch.conv1d((target.real*target.imag).to(weight.dtype)*self.dt, d_weight, padding='same')
        return filt_source, filt_target

Decoder.register('complex_conv', Complex_Spike_Conv)

class Spike_Conv(Decoder):
    def __init__(self, num=None, dec_target=None, dt=None, coding_method='spike_counts', coding_var_name='O', node_type=('excitatory', 'inhibitory', 'pyramidal', '...'), decay=0.9, ocillate=-0.01, **kwargs):
        super(Spike_Conv, self).__init__(num, dec_target, dt, coding_method, coding_var_name, node_type, **kwargs)
        self.kernel = kwargs.get('kernel', None)
        self.time_window = kwargs.get('time_window', 80)/2.0

        if self.kernel is None:
            tt = torch.arange(-self.time_window, self.time_window, self.dt)
            self.kernel = torch.exp(-(3*tt/self.time_window)**2).view(1, 1, -1)
        self.klen = self.kernel.shape[-1]

    def torch_coding(self, source, target, device='cpu'):
        weight = self.kernel.to(device).expand(source.shape[1],source.shape[1],self.klen)
        source = source.transpose(0, 1).transpose(1, 2)
        target = target.transpose(0, 1).transpose(1, 2)
        filt_source = torch.conv1d(source.to(weight.dtype), weight, padding='same')
        filt_target = torch.conv1d(target.to(weight.dtype), weight, padding='same')
        return filt_source, filt_target

Decoder.register('spike_conv', Spike_Conv)


# -*- coding: utf-8 -*-
"""
Created on 2020/8/17
@project: SPAIC
@filename: Network_saver
@author: Mengxiao Zhang
@contact: mxzhangice@gmail.com

@description:
对既定格式网络的存储
"""

import os
from spaic.Network.Assembly import Assembly
from spaic.Neuron.Neuron import NeuronGroup
from spaic.Neuron.Node import Node
from spaic.Network.Topology import Connection
from spaic.Backend.Backend import Backend
from spaic.Network.Topology import Projection
from spaic.Monitor.Monitor import Monitor
from spaic.IO.Initializer import BaseInitializer
from spaic.IO import Initializer as Initer

import time

def network_save(Net: Assembly, filename=None, path=None,
                 trans_format='json', combine=False, save=True, save_weight=True):
    '''
        Save network to files.

        Args:
            Net(Assembly) : The network needed to be saved.
            filename(Str) : The filename of the file that save target network.
            trans_format(str) : The format of file, could be 'json' or 'yaml'
            combine(Boolen) : Whether combine weight and structure of the Network into on file, False by default.
            save(Boolen) : Whether need to save the structure.

        Return:
            filename(str) : The filename of the file we save network, since it
                will give an auto name if no name given.

        Examples:
            >>> save_file = network_save(Net, "TestNetwork", trans_format='yaml', combine=True, save=True)
            ("TestNetwork", a dict of the structure of the network)
            or
            >>> save_file = network_save(Net, "TestNetwork", trans_format='json', combine=True, save=False)
            a dict of the structure of the network

    '''
    import numpy as np
    if filename is None:
        if Net.name:
            filename = Net.name + str(np.random.randint(10000))
        else:
            filename = "autoname" + str(np.random.randint(10000))

    origin_path = os.getcwd()
    if path:
        filedir = path + '/' + filename
    else:
        path = './'
        filedir = path + filename

    if save:
        os.chdir(path)
        if filename not in os.listdir():
            os.mkdir(filename)
        # os.chdir(filedir)
        os.chdir(filename)

    result_dict = trans_net(Net=Net, path=path, combine=combine, save=save, save_weight=save_weight)

    if save:
        if trans_format == "yaml":
            import yaml
            result = yaml.dump(result_dict, indent=4)
            ends = '.yml'
        elif trans_format == 'json':
            import json
            result = json.dumps(result_dict, indent=4)
            ends = '.json'
        else:
            raise ValueError("Wrong data format. Only support yaml/json format.")

        with open('./'+filename+ends, 'w+') as f:
            f.write(result)
        os.chdir(origin_path)
        print("Save Complete.")
        return filename
    else:
        os.chdir(origin_path)
        print("Complete.")
        return result_dict

def trans_net(Net: Assembly, path: str, combine: bool, save: bool, save_weight: bool):
    '''
        Transform the structure of the network for saving.

        Args:
            Net(Assembly): target network.
            path(string): Target path for saving net data.

        return:
            result_dict(dictionary) : the result diction of the whole Network.

        Example:
            yaml_net = trans_net(Net)

    '''
    result_dict = dict()
    net_name = Net.name
    result_dict[net_name] = []

    for g in Net._groups.values():
        if g._class_label == '<asb>':  # translate other assemblies
            sub_net_name = g.name
            result_dict[net_name].append(trans_net(g, path+'/'+str(sub_net_name),
                                                   combine, save, save_weight=False))
        elif g._class_label == '<neg>':  # translate layers
            result_dict[net_name].append(trans_layer(g))
        elif g._class_label == '<nod>':  # translate nodes
            result_dict[net_name].append(trans_node(g))
        else:
            # TODO: if get wrong _class_label, need check the type of
            #  this element
            pass

    for p in Net._projections.values():
        result_dict[net_name].append(trans_projection(p, combine, save_weight))

    for c in Net._connections.values():  # translate connections
        result_dict[net_name].append(trans_connection(c, combine, save_weight))

    if '_monitors' in dir(Net):
        mon_dict = {'monitor': []}
        result_dict[net_name].append(mon_dict)
        for monitor in Net._monitors.items():
            mon_dict['monitor'].append(trans_monitor(monitor))

    if '_learners' in dir(Net):
        for key, g in Net._learners.items():  # translate learners
            result_dict[net_name].append({key: trans_learner(g, key)})
    # result_dict[net_name].append({'learners':trans_learner(Net._learners)})

    if (not combine) and save_weight:
        if Net._backend:
            result_dict[net_name].append(
                {'backend': trans_backend(Net._backend, save)}
            )
        else:
            import warnings
            warnings.warn("Net._backend not exist. Please check whether need save weight")

    return result_dict


def trans_node(node: Node):
    '''
        Transform the structure of the Node layer for saving and extract the
            parameters.

        Args:
            node (Node): target node layer, like input layer and output layer

        return:
            result_dict (dictionary): the result diction with necessary
                parameters of the layer.

    '''

    needed = ['id', 'shape', 'num', '_time', '_dt', 'coding_method',
              'coding_var_name', 'type', 'name', 'coding_param']

    result_dict = dict()
    para_dict = dict()

    for key, para in node.__dict__.items():
        if key in needed:
            para_dict[key] = check_var_type(para)

    para_dict['shape'] = list(para_dict['shape'][1:])
    if node.is_encoded:
        para_dict['shape'] = para_dict['shape'][1:]

    if 'dt' in dir(node):
        para_dict['dt'] = node.dt
    if 'time' in dir(node):
        para_dict['time'] = node.time

    if node.__dict__['dec_target']:
        para_dict['dec_target'] = node.__dict__['dec_target'].name

    para_dict['kind'] = node._node_sub_class

    # if 'action' in node.__dict__.keys():
    #     para_dict['kind'] = 'Action'
    # elif 'reward' in node.__dict__.keys():
    #     para_dict['kind'] = 'Reward'
    # elif 'gen_first' in node.__dict__.keys():
    #     para_dict['kind'] = 'Generator'
    # elif 'predict' in node.__dict__.keys():
    #     para_dict['kind'] = 'Decoder'
    # else:
    #     para_dict['kind'] = 'Encoder'

    para_dict['_class_label'] = '<nod>'
    result_dict[node.name] = para_dict
    return result_dict


def trans_layer(layer: NeuronGroup):
    '''
        Transform the structure of the layer for saving and extract the
            parameters.

        Args:
            layer (NeuronGroup): target layer

        return:
            result_dict (dictionary): the result diction with necessary
                parameters of the layer.

    '''
    result_dict = dict()
    para_dict = dict()


    unneeded = [ 'enabled', '_backend', '_connections', '_supers', '_input_connections',
                '_output_connections', '_var_names', 'model_class', '_operations', 'model',
                '_groups']
    needed = ['model_name', 'id', 'name', 'num', 'position', 'shape', 'type', 'parameters']
    # Needed parameters: num, shape, neuron_type,
    # neuron_position, model, name, parameters.

    for key, para in layer.__dict__.items():
        if key in needed:
            para_dict[key] = check_var_type(para)

    if para_dict['position'] != ('x, y, z' or 'x, y'):
        para_dict.pop('position')
    para_dict['_class_label'] = '<neg>'

    para_dict['parameters'] = layer.parameters

    result_dict[layer.name] = para_dict
    return result_dict


def trans_projection(projection: Projection, combine: bool, save_weight: bool ):
    '''
        Transform the structure of the projection for saving and extract the
            parameters.

        Args:
            projection (Projection): target projection

        return:
            result_dict (dictionary): the result diction with necessary
                parameters of the projection.

    '''
    result_dict = dict()
    para_dict = dict()
    name_needed = ['pre', 'post']
    needed = ['name', 'link_type', 'ConnectionParameters']

    for key, para in projection.__dict__.items():
        if key in name_needed:
            para_dict[key] = para.name
        elif key in needed:
            para_dict[key] = check_var_type(para)

    # para_dict['prjs'] = {}
    para_dict['conns'] = []

    # for key, prj in projection._projections.items():
    #     para_dict['prjs'][key] = trans_projection(prj, combine, save_weight)
    for key, conn in projection._connections.items():
        para_dict['conns'].append(trans_connection(conn, combine, save_weight))



    para_dict['_class_label'] = '<prj>'
    # para_dict['_policies'] = []
    # for ply in projection._policies:
    #     if ply.name == 'Index_policy':
    #         para_dict['_policies'].append(
    #             {'name': ply.name,
    #              'pre_indexs': ply.pre_indexs,
    #              'post_indexs': ply.post_indexs,
    #              'level': ply.level}
    #         )
    #     else:
    #         para_dict['_policies'].append(
    #             {'name': ply.name,
    #              'pre_types': list(ply.pre_types) if ply.pre_types else ply.pre_types,
    #              'post_types': list(ply.post_types) if ply.post_types else ply.post_types,
    #              'level': ply.level}
    #         )

    result_dict[projection.name] = para_dict

    return result_dict


def trans_connection(connection: Connection, combine: bool, save_weight: bool):
    '''
        Transform the structure of the connection for saving and extract the
            parameters.

        Args:
            connection (Connection): target connection
            combine (bool): whether combine weights.

        return:
            result_dict (dictionary): the result diction with necessary
                parameters of the connection.

    '''
    result_dict = dict()
    para_dict = dict()

    name_needed = ['pre', 'post']
    needed = ['name', 'link_type', 'synapse_type', 'max_delay', 'sparse_with_mask',
              'pre_var_name', 'post_var_name', 'parameters', 'id', ]
    unneeded = ['hided', 'pre_groups', 'post_groups', 'pre_assemblies', 'post_assemblies',
                'unit_connections', '_var_names', '_supers', '_backend']
    # **link_parameters

    for key, para in connection.__dict__.items():
        if key in name_needed:
            para_dict[key] = para.id
        elif key in needed:
            d_para = para
            if key == 'parameters':
                if 'weight' in d_para.keys():
                    del d_para['weight']
                if 'bias' in d_para.keys():
                    d_para['bias'] = trans_bias(d_para['bias'])
            para_dict[key] = check_var_type(d_para)
    if combine:     # 是否需要在文件中存储weight
        para_dict['weight'] = check_var_type(connection.weight.value)

    para_dict['_class_label'] = '<con>'
    result_dict[connection.name] = para_dict

    return result_dict


def trans_backend(backend: Backend, save: bool):
    '''
    Transform the data of backend for saving.

    Args:
        backend: target backend.
        path(string): Target path for saving net data.

    Returns:
        result(dict): Contain the parameters of backend to be saved.
    '''

    # Needed parameters: _variables, _parameters_dict, _InitVariables_dict,
    # dt, time, time_step, _graph_var_dicts,

    # key_parameters_dict = ['_variables', '_parameters_dict', '_InitVariables_dict']
    key_parameters_dict = ['_parameters_dict']
    key_parameters_list = ['dt', 'runtime', 'time', 'n_time_step']


    if backend._variables is None:
        import warnings
        warnings.warn('Backend end don\'t have variables. Have not built Backend. Weight not exists.')
        return
    else:
        if 'parameters' not in os.listdir():
            os.mkdir('parameters')
    ori_path = os.getcwd()
    sim_path = ori_path + '/parameters'
    os.chdir(sim_path)

    import torch

    result_dict = dict()
    for key in key_parameters_dict:
        if save:
            save_path = sim_path + '/' + key + '.pt'
            data = backend.__dict__[key]
            torch.save(data, save_path)
            result_dict[key] = './parameters/' + key + '.pt'
        else:
            result_dict = backend._parameter_dict
            # pass
            # raise ValueError("Wrong save choosen, since parameters can be get from network"
            #                  "unneeded to use network_save function.")

    for key in key_parameters_list:
        result_dict[key] = backend.__dict__[key]

    result_dict['data_type'] = str(backend.__dict__['data_type'])

    os.chdir(ori_path)
    return result_dict


def trans_learner(learner, learn_name):
    """
    Transform learner parameters to dict.
    Args:
        learner: Target learner with needed parameters.

    Returns:
        result(dict): Contain the parameters of learner to be saved.
    """
    import torch
    # result_dict = dict()
    para_dict = dict()
    trainables = ['trainable_connections', 'trainable_groups', 'trainable_nodes']
    para_dict['trainable'] = []
    needed = ['name', 'parameters', 'optim_name', 'optim_lr', 'optim_para', 'lr_schedule_name', 'lr_schedule_para']
    para_dict['_class_label'] = '<learner>'
    for key in needed:
        if key in learner.__dict__.keys():
            para = learner.__dict__.get(key)
            para_dict[key] = check_var_type(para)
                # if type(para) != torch.Tensor \
                #     else para.detach().cpu().numpy().tolist()

    for train_name in trainables:
        for key, train in learner.__dict__[train_name].items():
            para_dict['trainable'].append(check_var_type(train.name))

    para_dict['algorithm'] = check_var_type(para_dict['name'])
    para_dict['name'] = check_var_type(learn_name)
    if 'algorithm' in para_dict['parameters'].keys():
        del para_dict['parameters']['algorithm']

    return para_dict


def trans_monitor(monitor: Monitor):
    """
    Transform monitor to dict.
    Args:
        learner: Target learner with needed parameters.

    Returns:
        result(dict): Contain the parameters of learner to be saved.
    """
    from spaic.Monitor.Monitor import StateMonitor, SpikeMonitor
    needed = ['var_name', 'index', 'dt', 'get_grad', 'nbatch']
    name, mon = monitor
    result_dict = dict()
    for i in needed:
        result_dict[i] = mon.__dict__[i]
    result_dict['target'] = mon.target.id
    result_dict['monitor_type'] = 'StateMonitor' if type(monitor[1]) == StateMonitor else 'SpikeMonitor'

    return {name: result_dict}


def check_var_type(var):
    import torch
    import numpy as np
    import json
    try:
        json.dumps(var)
        return var
    except:
        if isinstance(var, torch.Tensor):
            return var.detach().cpu().numpy().tolist()
        if isinstance(var, dict):
            for key, value in var.items():
                var[key] = check_var_type(value)
            return var
        if isinstance(var, str):
            return var
        try:
            var_list = var.tolist()
            return var_list
        except:
            raise TypeError('Please check type of parameters, we only support tensor or python build-in types.')
        # if len(var) >= 2:
        #     res_list = var.tolist()
        #     return [check_var_type(i) for i in res_list]
        # else:
        #     return check_var_type(var.tolist()[0])
    # if isinstance(var, torch.Tensor) or isinstance(var, np.ndarray):
    #     if len(var) >= 2:
    #         return var.tolist()
    #     else:
    #         return var.tolist()[0]
    # # elif isinstance(var, np.ndarray):
    # #     if len(var) >= 2:
    # #         return var.tolist()
    # #     else:
    # #         return var.tolist()[0]
    # elif isinstance(var, set):
    #     return list(var)
    # elif isinstance(var, list):  # 如果出现list中还有别的类型，考虑去后端解决
    #     return var
    # elif isinstance(var, dict):
    #     for key, value in var.items():
    #         var[key] = check_var_type(value)
    #     return var
    # else:
    #     return var


def trans_bias(para: dict):
    if isinstance(para, BaseInitializer):
        n_para = dict()
        for key in Initer.__all__:
            if Initer.__dict__[key] == para.__class__:
                n_para['method'] = key
                break
        n_para['para'] = para.__dict__
        return n_para
    else:
        return para









# -*- coding: utf-8 -*-
"""
Created on 2020/8/17
@project: SPAIC
@filename: Network_loader
@author: Mengxiao Zhang
@contact: mxzhangice@gmail.com

@description:
对已按格式储存网络的加载和重建
"""

import yaml
import json

from spaic.Network.Network import Network
from spaic.Network.Assembly import Assembly
from spaic.Neuron.Neuron import NeuronGroup
from spaic.Neuron.Node import Node, Decoder, Encoder, Action, Generator, Reward
from spaic.Network.Topology import Connection
from spaic.Backend.Backend import Backend
from spaic.Backend.Torch_Backend import Torch_Backend
from spaic.Learning.Learner import Learner
from spaic.Network.Topology import Projection
from spaic.Monitor.Monitor import Monitor, StateMonitor, SpikeMonitor
from spaic.IO.Initializer import BaseInitializer
from spaic.IO import Initializer as Initer

import torch


def network_load(filename=None, path=None, device='cpu', load_weight=True):
    '''
        The main function for getting the target filename and reloading the
            network.

        Args:
            filename(str) : The filename of the target network, given by user.
            dataloader(dataloader) : The dataloader for input node layer,
                should be given or crash.
            encoding(str) : The encoding model chosen by user.
            device(str) : The device type we choose to run our network.

        Return:
            net(Assembly) :The network that reloaded from the file.

        Example:
            Net = network_load('TestNetwork', dataloader, 'poisson')

    '''
    import os
    if path:
        filedir = path + '/' + filename
    else:
        path = './'
        # filedir = path + filename
    file = filename.split('.')[0]
    origin_path = os.getcwd()
    os.chdir(path+'/'+file)
    if os.path.exists(filename):
        with open(filename, 'r') as f:
            data = f.read()
            if data.startswith('{'):
                data = json.loads(data)
            else:
                data = yaml.load(data, Loader=yaml.FullLoader)

    else:
        if os.path.exists(filename+'.yml'):
            with open(filename+'.yml', 'r') as f:
                data = yaml.load(f, Loader=yaml.FullLoader)

        elif os.path.exists(filename+'.json'):
            with open(filename+'.json', 'r') as f:
                data = json.load(f)

        elif os.path.exists(filename+'.txt'):
            with open(filename+'.txt', 'r') as f:
                data = f.read()
                if data.startswith('{'):
                    data = json.loads(data)
                else:
                    data = yaml.load(data, Loader=yaml.FullLoader)

        else:
            raise ValueError("file %s doesn't exist, please check the "
                             "filename" % filename)

    net = ReloadedNetwork(net_data=data, device=device, load_weight=load_weight)

    os.chdir(origin_path)
    return net


class ReloadedNetwork(Network):
    '''
        The network rebuild from the yaml file.

        Args:
            net_data(dict) : The network information reloaded from yaml files.
            dataloader(dataloader) : The dataloader of input layer, since the
                large scale of data, we will not save the data.
            encoding(str) : The encoding model, default as poisson, will change
                in the future.
            backend(backend) : Backend that user want to use.
            learner(str) : The learning model of this network, will change in
                the future.
            learner_alpha(int) : The parameter alpha for learning model, will
                change in the future.
            device(str) : The type of device that run our model.

        Methods:
            load_net(self, data: dict) : The function for load the whole
                network, main function of this class.
            load_layer(self, layer: dict) : The function for load layer.
            load_connection(self, con: dict) : The function for load
                connection.
            load_node(self, node: dict) : The function for load node like input
                or output.
            load_backend(self, path: str): The function for load backend.

        Example:
            Net = ReloadNetwork(net_data, dataloader, 'poisson', backend,
                'STCA', 0.5)

    '''
    def __init__(self, net_data: dict, backend=None, device='cpu', load_weight=True, data_type=None):
        super(ReloadedNetwork, self).__init__()

        self.device = device
        self.name = list(net_data)[0]
        self._backend_info = []

        self.load_net(net_data)

        self.load_backend(backend, device=device, load_weight=load_weight, data_type=data_type)

        # self._learner = Learner(algorithm='STCA', lr=0.5, trainable=self)

        del self._backend_info

    def load_net(self, data):
        '''
            The function for load the whole network, main function of this class.

            Args:
                data(dict) : The data should contains the network structure and
                    parameter from yaml.

        '''
        setid = 0
        data = data[list(data)[0]]
        for g in data:
            if list(g)[0] == 'monitor':
                monitors = g.get('monitor')
                for monitor in monitors:
                    self.load_monitor(monitor)
                continue
            if list(g)[0] == 'backend':
                self._backend_info = g[list(g)[0]]
                continue
            para = g[list(g)[0]]
            if type(para) is dict:
                if para.get('_class_label') == '<neg>':
                    lay_name = para.get('name')
                    self.add_assembly(name=lay_name,
                                      assembly=self.load_layer(para))
                elif para.get('_class_label') == '<nod>':
                    nod_name = para.get('name')
                    self.add_assembly(name=nod_name,
                                      assembly=self.load_node(para))
                elif para.get('_class_label') == '<con>':
                    con_name = para.get('name')
                    self.add_connection(name=con_name,
                                        connection=self.load_connection(pnet=self, con=para))
                elif para.get('_class_label') == '<prj>':
                    prj_name = para.get('name')
                    self.add_projection(name=prj_name,
                                        projection=self.load_projection(prj=para))
                elif para.get('_class_label') == '<learner>':
                    learner = self.load_learner(para)
                    self._learners[para.get('name')] = learner
                else:
                    print('Unknown class label %d' % para['_class_label'])

                    break
            else:
                self.add_assembly(name=list(g)[0], assembly=self.load_assembly(list(g)[0], para))


    def load_assembly(self, name, assembly: list):
        target = Assembly(name=name)
        for g in assembly:
            para = g[list(g)[0]]
            if para.get('_class_label') == '<neg>':
                lay_name = para.get('name')
                target.add_assembly(name=lay_name,
                                  assembly=self.load_layer(para))
            elif para.get('_class_label') == '<nod>':
                nod_name = para.get('name')
                target.add_assembly(name=nod_name,
                                  assembly=self.load_node(para))
            elif para.get('_class_label') == '<con>':
                con_name = para.get('name')
                target.add_connection(name=con_name,
                                    connection=self.load_connection(pnet=target, con=para))
            elif para.get('_class_label') == '<prj>':
                prj_name = para.get('name')
                target.add_projection(name=prj_name,
                                    projection=self.load_projection(prj=para))
        return target

    def load_layer(self, layer: dict):
        '''
            The function for load layer.

            Args:
                layer(dict): Data contains the parameters of layers.

            Return：
                NeuronGroup with need parameters.

        '''
        # layer.pop('_class_label')
        # parameters = self.trans_para(layer.get('parameters'))
        parameters = layer.get('parameters')
        return_neuron = NeuronGroup(
            num= layer.get('num', 100),
            shape= layer.get('shape', [100]),
            neuron_type     = layer.get('type', 'non_type'),
            neuron_position = layer.get('position', 'x, y, z'),
            model= layer.get('model_name', 'clif'),
            name            = layer.get('name'),
            **parameters
        )
        return_neuron.id = layer.get('id', None)
        return return_neuron

    def load_connection(self, pnet, con: dict):
        '''
            The function for load connections,

            Args:
                con(dict): Data contains the parameters of connections.

            Return:
                Connection with needed parameters.

        '''

        if pnet._class_label == '<prj>':
            for pretarget in pnet.pre.get_groups():
                if con['pre'] == pretarget.id:
                    con['pre'] = pretarget
            for posttarget in pnet.post.get_groups():
                if con['post'] == posttarget.id:
                    con['post'] = posttarget
        else:
            for target in pnet.get_groups():
                if con['pre'] == target.id:
                    con['pre'] = target
                if con['post'] == target.id:
                    con['post'] = target

        for con_tar in ['pre', 'post']:
            if isinstance(con[con_tar], str):
                con[con_tar] = self.get_elements()[con[con_tar]] if con[con_tar] in self.get_elements().keys() else None

        assert not isinstance(con['pre'], str)
        assert not isinstance(con['post'], str)

        if 'bias' in con['parameters'].keys():
            bias = con['parameters']['bias']
            if isinstance(con['parameters']['bias'], dict):
                if 'method' in con['parameters']['bias'].keys():
                    method = bias.get('method')
                    con['parameters']['bias'] = Initer.__dict__[method](**bias.get('para'))


        # con.pop('weight_path')
        return_conn = Connection(
            pre              = con.get('pre'),
            post             = con.get('post'),
            name             = con.get('name'),
            link_type        = con.get('link_type', 'full'),
            syn_type         = con.get('synapse_type', ['basic_synapse']),
            max_delay        = con.get('max_delay', 0),
            sparse_with_mask = con.get('sparse_with_mask', False),
            pre_var_name     = con.get('pre_var_name', 'O'),
            post_var_name    = con.get('post_var_name', 'WgtSum'),
            **con.get('parameters')
        )
        return_conn.id = con.get('id', None)
        return return_conn

    def load_projection(self, prj: dict):
        '''
            The function for load projection,

            Args:
                prj(dict): Data contains the parameters of projection.

            Return:
                Projection with needed parameters.

        '''
        if prj['pre'] in self._groups.keys() and \
                prj['post'] in self._groups.keys():
            prj['pre']  = self._groups[prj['pre']]
            prj['post'] = self._groups[prj['post']]
        else:
            print("Trans_error")
            print(self._groups.keys())

        assert not isinstance(prj['pre'], str)
        assert not isinstance(prj['post'], str)

        this_prj = Projection(
            pre                  = prj.get('pre'),
            post                 = prj.get('post'),
            name                 = prj.get('name'),
            link_type            = prj.get('link_type', 'full'),
            # policies             = prj.get('policies', []),
            ConnectionParameters = prj.get('ConnectionParameters'),
        )

        for conn in prj['conns']:
            for key, value in conn.items():
                this_prj.add_connection(
                    con=self.load_connection(pnet=this_prj, con=value),
                    name=key,
                )


        # prj['policies'] = []
        # from Network.ConnectPolicy import IndexConnectPolicy, ExcludedTypePolicy, IncludedTypePolicy
        # policy_dict = {'Included_policy': IncludedTypePolicy,
        #                'Excluded_policy': ExcludedTypePolicy}
        #
        # for ply in prj['_policies']:
        #     if ply['name'] == 'Index_policy':
        #         prj['policies'].append(IndexConnectPolicy(pre_indexs=ply['pre_indexs'],
        #                                                   post_indexs=ply['post_indexs'],
        #                                                   level=ply['level']))
        #     else:
        #         prj['policies'].append(policy_dict[ply['name']](pre_types=ply['pre_types'],
        #                                                   post_types=ply['post_types'],
        #                                                   level=ply['level']))

        # con.pop('weight_path')
        return this_prj

    def load_node(self, node: dict):
        '''
            The function for load node like input or output.

            Args:
                node(dict): Data contains the parameters of nodes.

            Return:
                Node of input or output layer, contains needed parameters.

        '''

        Node_dict = {'<decoder>': Decoder, '<action>': Action, '<reward>': Reward,
                     '<generator>': Generator, '<encoder>': Encoder}

        if node.get('kind') == '<decoder>':
            return_node = Node_dict[node.get('kind')](
                num             = node.get('num'),
                dec_target      = self._groups.get(node.get('dec_target', None), None),
                dt              = node.get('dt', 0.1),
                # time            = node.get('time'),
                coding_method   = node.get('coding_method', 'poisson'),
                coding_var_name = node.get('coding_var_name', 'O'),
                node_type       = node.get('type', None),
                **node.get('coding_param')
            )
        else:
            return_node = Node_dict[node.get('kind')](
                shape           = node.get('shape', None),
                num             = node.get('num'),
                dec_target      = self._groups.get(node.get('dec_target', None), None),
                dt              = node.get('dt', 0.1),
                # time            = node.get('time'),
                coding_method   = node.get('coding_method', 'poisson'),
                coding_var_name = node.get('coding_var_name', 'O'),
                node_type       = node.get('type', None),
                **node.get('coding_param')
            )
        return_node.id = node.get('id', None)
        return return_node

    def load_backend(self, backend=None, device=None, load_weight=False, data_type=None):
        '''
            The function for load backend parameters.

        '''

        key_parameters_list = ['dt', 'runtime', 'time', 'n_time_step']
        typical = ['_graph_var_dicts']

        import torch

        if backend is None:
            backend = Torch_Backend(device)
        self.set_backend(backend)
        self.set_backend_data_type(data_type=data_type)
        if self._backend_info:
            for key in key_parameters_list:
                self._backend.__dict__[key] = self._backend_info[key]
        self.build()

        if load_weight:
            path = self._backend_info['_parameters_dict']
            data = torch.load(path, map_location=self._backend.device0)
            for key, value in data.items():
                # print(key, 'value:', value)
                if key in self._backend._parameters_dict.keys():
                    target_device = self._backend._parameters_dict[key].device
                else:
                    target_device = self._backend.device0
                self._backend._parameters_dict[key] = value.to(target_device)

        return

    def set_backend_data_type(self, data_type=None):
        import torch
        supported_data_type = {'torch.float64': torch.float64,
                               'torch.float32': torch.float32,
                               'torch.float16': torch.float16,
                               'torch.bfloat16': torch.bfloat16,
                               'torch.int64': torch.int64,
                               'torch.int32': torch.int32,
                               'torch.int16': torch.int16,
                               'torch.bool': torch.bool,
                               'torch.uint8': torch.uint8}
        if data_type:
            self._backend.data_type = data_type
        else:
            if self._backend_info:
                self._backend.data_type = supported_data_type[self._backend_info['data_type']]
            else:
                self._backend.data_type = supported_data_type['torch.float32']

    def load_learner(self, learner: dict):
        '''
            The function for load learners' parameters.

        '''
        if '<net>' in learner['trainable']:  ## If self in net, use the whole net as the trainable target.
            learner.pop('trainable')
            builded_learner = Learner(
                algorithm = learner.get('algorithm'),
                trainable = self,
                **learner.get('parameters')
            )
        else:
            trainable_list = []
            for trains in learner['trainable']:
                if trains in self._groups:
                    trainable_list.append(self._groups[trains])
                elif trains in self._connections:
                    trainable_list.append(self._connections[trains])
            learner.pop('trainable')
            if learner.get('parameters'):
                builded_learner = Learner(
                    trainable = trainable_list,
                    algorithm = learner.get('algorithm'),
                    **learner.get('parameters')
                    )
            else:
                builded_learner = Learner(
                    trainable = trainable_list,
                    algorithm = learner.get('algorithm')
                )
        if learner.get('optim_name', None):
            builded_learner.set_optimizer(optim_name=learner.get('optim_name'),
                                      optim_lr=learner.get('optim_lr'),
                                      **learner.get('optim_para'))
        if learner.get('lr_schedule_name', None):
            builded_learner.set_schedule(lr_schedule_name=learner.get('lr_schedule_name'),
                                         **learner.get('lr_schedule_para'))

        return builded_learner

    def load_monitor(self, monitor: dict):
        '''
        Used to add monitors to the model according to the

        Args:
            monitor: a dict that contains monitors' information.


        '''
        monitor_dict = {'StateMonitor': StateMonitor,
                        'SpikeMonitor': SpikeMonitor}

        for name, mon in monitor.items():
            for target in self.get_groups():
                if mon['target'] == target.id:
                    mon['target'] = target
                    break
            for target in self.get_connections():
                if mon['target'] == target.id:
                    mon['target'] = target
                    break

            self.add_monitor(name=name,
                                 monitor=monitor_dict[mon.get('monitor_type', 'StateMonitor')](
                                 target=mon['target'],
                                 var_name=mon['var_name'],
                                 dt=mon['dt'],
                                 get_grad=mon['get_grad'],
                                 nbatch=mon['nbatch'],
                                 index=mon['index']))

    def trans_para(self, para):
        if isinstance(para, dict):
            for key, value in para.items():
                para[key] = self.trans_para(value)
        else:
            para = torch.tensor(para, dtype=torch.float32, device=self.device[0])
        return para